import argparse
import logging

from cardutil.cli import add_version, print_banner
from cardutil.mciipm import IpmReader, IpmWriter


def cli_entry():
    cli_run(**vars(cli_parser().parse_args()))


def cli_run(**kwargs):
    print_banner('mci_ipm_encode', kwargs)

    if kwargs.get('debug'):
        logging.basicConfig(level=logging.DEBUG)

    if not kwargs.get('out_filename'):
        kwargs['out_filename'] = kwargs['in_filename'] + '.out'

    # if --no1014encoding provided, override new in-format and out-format parms
    if kwargs.get('no1014blocking'):
        kwargs['in_format'] = kwargs['out_format'] = 'vbs'

    with open(kwargs['in_filename'], 'rb') as in_file, open(kwargs['out_filename'], 'wb') as out_file:
        mci_ipm_encode(in_file, out_file=out_file, **kwargs)


def cli_parser():
    parser = argparse.ArgumentParser(prog='mci_ipm_encode', description='Mastercard IPM file encoder')
    parser.add_argument('in_filename')
    parser.add_argument('-o', '--out-filename')
    parser.add_argument('--in-encoding')
    parser.add_argument('--out-encoding')
    parser.add_argument('--no1014blocking', action='store_true')
    parser.add_argument('--in-format', choices=['vbs', '1014'], default='1014')
    parser.add_argument('--out-format', choices=['vbs', '1014'], default='1014')
    parser.add_argument('--debug', action='store_true')
    add_version(parser)
    return parser


def mci_ipm_encode(in_file, out_file=None, in_encoding='cp500', out_encoding='latin_1',
                   in_format='1014', out_format='1014', **_):
    """
    Change encoding of IPM file from one encoding scheme to another

    :param in_file: input IPM file object
    :param out_file: output IPM file object
    :param in_encoding: input file encoding string
    :param out_encoding: output file encoding string
    :param in_format: input file format (vbs/1014)
    :param out_format: output file format (vbs/1014)
    :return: None
    """

    in_blocked = True if in_format == '1014' else False
    with IpmWriter(out_file, encoding=out_encoding, blocked=in_blocked) as writer:
        out_blocked = True if out_format == '1014' else False
        reader = IpmReader(in_file, encoding=in_encoding, blocked=out_blocked)
        writer.write_many(reader)


if __name__ == '__main__':
    cli_entry()



import io
import unittest

from cardutil import CardutilError
from cardutil.mciipm import (
    VbsWriter, VbsReader, IpmReader, IpmWriter, Block1014, Unblock1014, block_1014, unblock_1014, vbs_list_to_bytes,
    vbs_bytes_to_list, IpmParamReader, MciIpmDataError)

from tests import message_ascii_raw, message_ebcdic_raw, print_stream


class MciIpmTestCase(unittest.TestCase):

    def test_mciipm_data_error_exception(self):
        err = MciIpmDataError("Message", record_number=1, binary_context_data=b'1234')
        assert str(err) == 'Message'
        assert err.record_number == 1
        assert err.binary_context_data == b'1234'

    def test_real_message_example_ascii(self):
        # create the input ipm file bytes -- test_file
        message_list = [message_ascii_raw for _ in range(15)]
        with io.BytesIO() as in_data:
            with VbsWriter(in_data, blocked=True) as writer:
                writer.write_many(message_list)

            print_stream(in_data, "VBS in data")

            # read vbs test file
            reader = IpmReader(in_data, blocked=True)
            results = list(reader)
        self.assertEqual(len(results), len(message_list))

    def test_ipm_reader_iso8583_exception(self):
        """
        Test what happens when the iso8583 rec raises an exemption
        """
        # create the input ipm file bytes -- test_file
        with io.BytesIO() as in_data:
            with VbsWriter(in_data, blocked=True) as writer:
                writer.write(b'bad ipm message')
            reader = IpmReader(in_data, blocked=True)
            with self.assertRaises(CardutilError):
                list(reader)

    def test_real_message_example_ebcdic(self):
        # write 1014 blocked test file
        message_list = [message_ebcdic_raw for _ in range(15)]
        with io.BytesIO() as in_data:
            with VbsWriter(in_data, blocked=True) as writer:
                writer.write_many(message_list)

            print_stream(in_data, "1014 blocked in data")

            # read blocked test file
            reader = IpmReader(in_data, encoding='cp500', blocked=True)
            results = list(reader)
            print(results)

        self.assertEqual(len(results), len(message_list))

    def test_ipmwriter_vbs_file(self):
        record = {'MTI': '1111', 'DE2': '8888999988889999'}
        records = [record for _ in range(5)]

        with io.BytesIO() as out_data:
            writer = IpmWriter(out_data)
            for record in records:
                writer.write(record)
            writer.close()

            print_stream(out_data, 'VBS output file')

            reader = IpmReader(out_data)
            results = list(reader)
            print(results)

        self.assertEqual(results, records)

    def test_ipm_reader_with_config(self):
        record = {'MTI': '1111', 'DE2': '8888999988889999'}
        # the following config applies the PAN masking formatter
        bit_config = {"2": {"field_name": "PAN", "field_type": "LLVAR", "field_length": 0, "field_processor": "PAN"}}
        records = [record]

        with io.BytesIO() as out_data:
            with IpmWriter(out_data) as writer:
                writer.write_many(records)

            print_stream(out_data, 'VBS output file')
            reader = IpmReader(out_data, iso_config=bit_config)

            results = list(reader)

        print(results)
        self.assertEqual(results, [{'MTI': '1111', 'DE2': '888899******9999'}])

    def test_ipmwriter_blocked_file(self):
        record = {'MTI': '1111', 'DE2': '8888999988889999'}
        records = [record for _ in range(5)]

        with io.BytesIO() as out_data:
            with IpmWriter(out_data, blocked=True) as writer:
                writer.write_many(records)

            print_stream(out_data, 'VBS output file')

            reader = IpmReader(out_data, blocked=True)
            results = list(reader)
            print(results)

        self.assertEqual(results, records)

    def test_vbsreader_vbs_file(self):
        # create the input file bytes -- test_file
        records = [b'12345678901234567890' for _ in range(5)]
        with io.BytesIO() as in_data:

            # write vbs test file
            writer = VbsWriter(in_data)
            for record in records:
                writer.write(record)
            else:
                writer.close()
            print_stream(in_data, "VBS in data")

            reader = VbsReader(in_data)
            results = list(reader)
            print(results)

        self.assertEqual(results, records)

    def test_vbsreader_vbs_file_missing_0_len(self):
        """
        The reader can handle VBS files that don't have final 0 length record
        """
        # create the input file bytes -- test_file
        records = [b'12345678901234567890' for _ in range(5)]
        with io.BytesIO() as in_data:

            # write vbs test file
            writer = VbsWriter(in_data)
            # don't call close method which writes the zero length record
            for record in records:
                writer.write(record)

            print_stream(in_data, "VBS in data")

            reader = VbsReader(in_data)
            results = list(reader)
            print(results)

        self.assertEqual(results, records)

    def test_vbsreader_blocked_file(self):
        # create the input file bytes -- test_file
        records = [b'12345678901234567890' for _ in range(5)]
        with io.BytesIO() as in_data:

            # write vbs test file
            writer = VbsWriter(in_data, blocked=True)
            for record in records:
                writer.write(record)
            else:
                writer.close()
            print_stream(in_data, "Blocked vbs data")

            reader = VbsReader(in_data, blocked=True)
            results = list(reader)
            print(results)
            self.assertEqual(results, records)

    def test_vbsreader_exceptions(self):
        # create the input file bytes -- test_file
        with io.BytesIO() as in_data:
            in_data.write(b'\xFF\xFF\x00\x00')  # negative length
            print_stream(in_data, "VBS in data")
            reader = VbsReader(in_data)
            with self.assertRaises(MciIpmDataError):
                list(reader)

        with io.BytesIO() as in_data:
            in_data.write(b'\x00\x00\x00\x05abcd')  # one byte short of record length (5)
            print_stream(in_data, "VBS in data")
            reader = VbsReader(in_data)
            with self.assertRaises(MciIpmDataError) as context:
                list(reader)
            assert str(context.exception) == "Unable to read complete record - record length: 5, data read: 4"
            assert context.exception.record_number == 1
            assert context.exception.binary_context_data == b"\x00\x00\x00\x05abcd"

    def test_file_blocker_compare(self):
        """
        Checks that the Block1014 class works the same as the
        :return:
        """
        out_unblocked = io.BytesIO()

        message_list = [message_ascii_raw for _ in range(10)]
        writer = VbsWriter(out_unblocked)
        for message in message_list:
            writer.write(message)
        writer.close()

        out_blocked = io.BytesIO()
        block_1014(out_unblocked, out_blocked)
        out_blocked.seek(0)
        blocked1 = out_blocked.read()
        print(blocked1)

        out_blocked2 = io.BytesIO()
        writer = VbsWriter(out_blocked2, blocked=True)
        for message in message_list:
            writer.write(message)
        writer.close()
        out_blocked2.seek(0)
        blocked2 = out_blocked2.read()
        print(blocked2)

        self.assertEqual(blocked1, blocked2)

        out_blocked.seek(0)
        out = io.BytesIO()
        unblock_1014(out_blocked, out)

        print_stream(out, "unblocked data")

    def test_unblock1014_exceptions(self):
        # create correct blocked
        message_list = [message_ascii_raw for _ in range(10)]
        out_blocked = io.BytesIO()
        writer = VbsWriter(out_blocked, blocked=True)
        for message in message_list:
            writer.write(message)
        writer.close()
        out_blocked.seek(0)
        blocked = out_blocked.read()
        print(blocked)

        # remove byte from end of file -- invalid file size
        out_blocked_missing_data = io.BytesIO(blocked[:-2])

        out_blocked_missing_data.seek(0)
        out = io.BytesIO()
        with self.assertRaises(CardutilError):
            unblock_1014(out_blocked_missing_data, out)

        # bad pad chars
        out_blocked_bad_fill = io.BytesIO(blocked[:-2] + b'$$')
        out_blocked_bad_fill.seek(0)
        out = io.BytesIO()
        with self.assertRaises(CardutilError):
            unblock_1014(out_blocked_bad_fill, out)

    def test_write_read_large_records(self):
        """
        Checks that the Block1014 class handles large records (greater than 1014 bytes per record)
        """
        blocked = io.BytesIO()

        message_list = [b'*' * 2000 for _ in range(5)]
        writer = VbsWriter(blocked, blocked=True)
        for message in message_list:
            writer.write(message)
        writer.close()
        print_stream(blocked, 'blocked')

        reader = VbsReader(blocked, blocked=True)
        for count, rec in enumerate(reader):
            self.assertLess(count, 5)
            self.assertEqual(rec, b'*' * 2000)

    def test_block1014_file_obj(self):
        """
        check that can access the underlying file object
        :return:
        """
        my_file = io.BytesIO()
        my_file_block = Block1014(my_file)
        self.assertEqual(my_file_block.tell(), 0)
        self.assertEqual(my_file_block.undefined_func, None)
        my_file_block.close()

    def test_vbsreader_file_obj(self):
        """
        check that can access the underlying file object
        :return:
        """
        my_file = io.BytesIO()
        vbs = VbsReader(my_file)
        self.assertEqual(vbs.tell(), 0)
        self.assertEqual(vbs.undefined_func, None)

    def test_vbswriter_file_obj(self):
        """
        check that can access the underlying file object
        :return:
        """
        my_file = io.BytesIO()
        vbs = VbsWriter(my_file)
        self.assertEqual(vbs.tell(), 0)
        self.assertEqual(vbs.undefined_func, None)

    def test_unblock1014_file_obj(self):
        """
        check that can access the underlying file object
        :return:
        """
        my_file = io.BytesIO()
        my_file_block = Unblock1014(my_file)
        self.assertEqual(my_file_block.tell(), 0)
        self.assertEqual(my_file_block.undefined_func, None)
        my_file_block.read()

    def test_vbs_list_to_bytes_to_list(self):
        test_bytes_list = [b'aaa', b'bbb', b'ccc']
        vbs_data = vbs_list_to_bytes(test_bytes_list)
        print(vbs_data)
        self.assertEqual(vbs_data, b'\x00\x00\x00\x03aaa\x00\x00\x00\x03bbb\x00\x00\x00\x03ccc\x00\x00\x00\x00')
        vbs_list = vbs_bytes_to_list(vbs_data)
        print(vbs_list)
        self.assertEqual(vbs_list, test_bytes_list)

    def test_ipm_param_reader(self):
        param_file_data = [
            b'2011101414AIP0000T1IP0000T1 TABLE LIST                 ' + 188 * b'.' + b'001',
            b'2014101414AIP0000T1IP0040T1 ACCOUNT RANGE TABLE        ' + 188 * b'.' + b'036',
            b'TRAILER RECORD IP0000T1  00000218                                               ',
            b'........xxx....',  # dummy record
            b'1711114A0365116545113000000000MCC5116545113999999999MCC020000000152710084563AUS036CMCC NNYMCC N0000000362'
            b'0000000000000000000000000000 000000NN   000000NNNN0NUNN0N N     ',
        ]

        with io.BytesIO() as test_param_stream:
            with VbsWriter(test_param_stream, blocked=True) as test_param_vbs:
                test_param_vbs.write_many(param_file_data)

            test_param_stream.seek(0)
            reader = IpmParamReader(test_param_stream, table_id='IP0040T1')

            for record in reader:
                print(record)


if __name__ == '__main__':
    unittest.main()



import csv
import io
import os
import tempfile
import unittest

from tests import print_stream
from cardutil.vendor import hexdump
from cardutil.cli import mci_csv_to_ipm, mci_ipm_to_csv
from cardutil.config import config
from cardutil.mciipm import IpmReader


class MciCsvToIpmParserTestCase(unittest.TestCase):
    def test_mci_ipm_encode_cli_parser(self):
        args = vars(mci_csv_to_ipm.cli_parser().parse_args(['file1.ipm']))
        self.assertEqual(
            args,
            {'in_filename': 'file1.ipm', 'in_encoding': None, 'out_filename': None, 'out_encoding': None,
             'no1014blocking': False, 'config_file': None, 'debug': False})


class MciCsvToIpmIOTestCase(unittest.TestCase):
    def test_mci_csv_to_ipm_to_csv(self):
        """
        create an ipm file from csv, then create csv from ipm. Check that input csv == output csv
        """
        def do_test(no_blocking):
            csv_data = 'MTI,DE2,DE4,DE12\n0100,1111222233334444,100,2020-06-18'
            in_csv = io.StringIO(csv_data)
            print_stream(in_csv, 'in_csv')
            in_csv.seek(0)
            out_ipm = io.BytesIO()

            mci_csv_to_ipm.mci_csv_to_ipm(in_csv=in_csv, out_ipm=out_ipm, no1014blocking=no_blocking, config=config)
            out_ipm.seek(0)
            print_stream(out_ipm, 'out_ipm')
            out_ipm.seek(0)

            out_csv = io.StringIO()
            mci_ipm_to_csv.mci_ipm_to_csv(in_ipm=out_ipm, out_csv=out_csv, config=config, no1014blocking=no_blocking)
            print_stream(out_csv, 'out_csv')
            out_csv.seek(0)

            reader = csv.DictReader(out_csv)
            record = dict(list(reader).pop())
            print(record)
            self.assertEqual(record['MTI'], '0100')
            self.assertEqual(record['DE2'], '1111222233334444')
            self.assertEqual(record['DE4'], '100')
            self.assertEqual(record['DE12'], '2020-06-18 00:00:00')
        do_test(no_blocking=False)
        do_test(no_blocking=True)


class MciCsvToIpmTestCase(unittest.TestCase):

    temp_filenames = []

    def setUp(self) -> None:
        self.temp_filenames = []

    def tearDown(self) -> None:
        for filename in self.temp_filenames:
            if os.path.exists(filename):
                os.remove(filename)

    def run_cli_with_csv(self, csv, **kwargs):
        with tempfile.NamedTemporaryFile(mode='w', delete=False) as in_csv:
            in_csv.write(csv)
            in_csv_name = in_csv.name
            print(in_csv_name)
            in_csv.close()
        mci_csv_to_ipm.cli_run(in_filename=in_csv_name, **kwargs)
        self.temp_filenames.append(in_csv_name)
        if 'out_filename' in kwargs:
            outfile = kwargs['out_filename']
        else:
            outfile = in_csv_name + '.ipm'
        self.temp_filenames.append(outfile)
        self.print_hexdump(outfile)
        return self.get_ipm_data(outfile)

    def print_hexdump(self, filename):
        with open(filename, 'rb') as outfile:
            output = outfile.read()
        hexdump.hexdump(output)

    def get_ipm_data(self, filename):
        with open(filename, 'rb') as outfile:
            reader = IpmReader(outfile, blocked=True)
            recs = [record for record in reader]
        return recs

    def test_csv_to_ipm_input_params(self):
        """
        Run mci_csv_to_ipm using real files
        :return:
        """
        in_csv_data = 'MTI,DE2,DE4\n0100,1111222233334444,100'
        # run the conversion
        recs = self.run_cli_with_csv(in_csv_data, out_encoding='latin1')
        # perform the checks
        self.assertEqual(len(recs), 1)  # one record returned
        rec = recs[0]
        print(rec)
        recs = self.run_cli_with_csv(in_csv_data, out_filename='x')

    def test_csv_to_ipm_with_output_filename(self):
        in_csv_data = 'MTI,DE2,DE4\n0100,1111222233334444,100'
        out_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
        recs = self.run_cli_with_csv(in_csv_data, out_filename=out_file.name)
        print(recs)

    def test_csv_to_ipm_with_debug(self):
        in_csv_data = 'MTI,DE2,DE4\n0100,1111222233334444,100'
        out_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
        recs = self.run_cli_with_csv(in_csv_data, out_filename=out_file.name, debug=True)
        print(recs)

    def test_csv_to_ipm_has_DE48_only(self):
        """
        Run mci_csv_to_ipm test with DE48 only
        In this case, DE written as is
        """
        in_csv_data = ('MTI,DE2,DE3,DE4,DE12,DE14,DE22,DE23,DE24,DE25,DE26,DE30,DE31,DE33,DE37,DE38,DE40,DE41,DE42,'
                       'DE48,DE49,DE50,DE63,DE71,DE73,DE93,DE94,DE95,DE100,DE43_NAME,DE43_SUBURB,DE43_POSTCODE,'
                       'ICC_DATA\n'
                       '1644,,,,,,,,697,,,,,,,,,,,01050250022112070000005167800005011002500221120700000015995000050'
                       '122001T,,,,00000001,,,,,,,,,')

        # run the conversion
        recs = self.run_cli_with_csv(in_csv_data, out_encoding='latin1', debug=True)
        # perform the checks
        self.assertEqual(len(recs), 1)  # one record returned
        rec = recs[0]
        print(rec)
        self.assertEqual(rec['DE48'], '01050250022112070000005167800005011002500221120700000015995000050122001T')

    def test_csv_to_ipm_has_DE48_and_PDS(self):
        """
        Run mci_csv_to_ipm test with DE48 column and PDS
        In this case, PDS overrides DE, and only PDS value placed into PDS fields, DE values dropped
        """
        in_csv_data = ('MTI,DE2,DE3,DE4,DE12,DE14,DE22,DE23,DE24,DE25,DE26,DE30,DE31,DE33,DE37,DE38,DE40,DE41,DE42,'
                       'DE48,DE49,DE50,DE63,DE71,DE73,DE93,DE94,DE95,DE100,DE43_NAME,DE43_SUBURB,DE43_POSTCODE,'
                       'ICC_DATA,PDS0191\n'
                       '1644,,,,,,,,697,,,,,,,,,,,010502500221120700000051678000050110025002211207000000159950000501'
                       '22001T,,,,00000001,,,,,,,,,,2')

        # run the conversion
        recs = self.run_cli_with_csv(in_csv_data, out_encoding='latin1')

        # perform the checks
        self.assertEqual(len(recs), 1)  # one record returned
        rec = recs[0]
        print(rec)
        self.assertEqual(rec['MTI'], '1644')
        self.assertEqual(rec['DE24'], '697')
        self.assertEqual(rec['DE71'], 1)
        # check that only PDS field loaded, prior DE48 values are dropped
        self.assertEqual(rec['DE48'], '01910012')

    def test_csv_to_ipm_has_PDS_only(self):
        """
        Run mci_csv_to_ipm test with PDS fields only
        Check that all provided PDS fields are loaded into DE fields
        :return:
        """
        in_csv_data = ('MTI,DE2,DE3,DE4,DE12,DE14,DE22,DE23,DE24,DE25,DE26,DE30,DE31,DE33,DE37,DE38,DE40,DE41,DE42,'
                       'DE49,DE50,DE63,DE71,DE73,DE93,DE94,DE95,DE100,DE43_NAME,DE43_SUBURB,DE43_POSTCODE,ICC_DATA,'
                       'PDS0191,PDS0122\n1644,,,,,,,,697,,,,,,,,,,,,,,00000001,,,,,,,,,,2,T')
        # run the conversion
        recs = self.run_cli_with_csv(in_csv_data, out_encoding='latin1')
        # perform the checks
        self.assertEqual(len(recs), 1)  # one record returned
        rec = recs[0]
        print(rec)
        # de48 contains all PDS values passed
        self.assertEqual(rec['DE48'], '0122001T01910012')


if __name__ == '__main__':
    unittest.main()



import binascii
from array import array


class BitArray:
    """
    This is a minimal native python replacement for the bitarray module that was used to interpret
    the iso binary bitmap. The bitarray module is written in c and the author does not provide binary wheels so
    forces users to have compilers installed just to install.
    This small class provides the required functions from that library written in pure python.
    """
    endian = 'big'
    bytes = b''

    def __init__(self, endian='big'):
        self.endian = endian

    def frombytes(self, array_bytes):
        self.bytes = array_bytes

    def tobytes(self):
        return self.bytes

    def tolist(self):
        swap_bytes = array('B', self.bytes)
        if self.endian == 'little':
            for i, n in enumerate(swap_bytes):
                swap_bytes[i] = int('{:08b}'.format(n)[::-1], 2)
        width = len(self.bytes)*8
        swapped_bytes = swap_bytes.tobytes()
        bit_list = '{bytes:0{width}b}'.format(bytes=int(binascii.hexlify(swapped_bytes), 16), width=width)
        return [bit == '1' for bit in bit_list]

    def fromlist(self, bytelist):
        # https://stackoverflow.com/questions/32675679/convert-binary-string-to-bytearray-in-python-3
        binary_value = ''.join(['1' if val else '0' for val in bytelist])
        self.bytes = int(binary_value, 2).to_bytes(len(binary_value) // 8, byteorder='big')



import os
import tempfile
import unittest

from cardutil.cli import get_config


class GetConfigTestCase(unittest.TestCase):
    def test_envvar_config(self):
        with tempfile.NamedTemporaryFile('w+', delete=False) as config_file:
            config_file.write('{"config": "hello"}')
            config_full_filename = config_file.name
            config_dir = os.path.dirname(config_file.name)
            config_filename = os.path.basename(config_file.name)
            config_file.close()
        os.environ['TEST_ENVVAR'] = config_dir
        config = get_config(config_filename, envvar="TEST_ENVVAR")
        print(config)
        os.remove(config_full_filename)
        if os.environ.get('TEST_ENVVAR'):
            del os.environ['TEST_ENVVAR']
        self.assertEqual(config, {"config": "hello"})

    def test_envvar_config_not_found(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            os.environ['TEST_ENVVAR'] = temp_dir
            config = get_config("test", envvar="TEST_ENVVAR")
        self.assertIn('bit_config', config.keys())
        print(config)

    def test_cli_config(self):
        with tempfile.NamedTemporaryFile('w+', delete=False) as config_file:
            config_file.write('{"config": "hello2"}')
            config_full_filename = config_file.name
            config_filename = os.path.basename(config_file.name)
            config_file.close()
        if os.environ.get('TEST_ENVVAR'):
            del os.environ['TEST_ENVVAR']
        config = get_config(config_filename, envvar="TEST_ENVVAR", cli_filename=config_full_filename)
        os.remove(config_full_filename)
        self.assertEqual(config, {"config": "hello2"})

    def test_cli_config_not_found(self):
        config = get_config("test", envvar="TEST_ENVVAR", cli_filename="THIS_FILE_DOES_NOT_EXIST")
        self.assertIn('bit_config', config.keys())


if __name__ == '__main__':
    unittest.main()



import binascii


def print_stream(stream, description):
    stream.seek(0)
    data = stream.read()
    print('***' + description + '***')
    print(data)
    stream.seek(0)


def test_message(encoding='ascii', hex_bitmap=False):
    binary_bitmap = b'\xF0\x10\x05\x42\x84\x61\x80\x02\x02\x00\x00\x04\x00\x00\x00\x00'
    bitmap = binary_bitmap
    if hex_bitmap:
        bitmap = binascii.hexlify(binary_bitmap)
    return (
        '1144'.encode(encoding) +
        bitmap +
        ('164444555544445555111111000000009999150815171500123456789012333123423579957991200000'
         '012306120612345612345657994211111111145BIG BOBS\\80 KERNDALE ST\\DANERLEY\\3103  VIC'
         'AUS0080001001Y99901600000000000000011234567806999999').encode(encoding))


message_ascii_raw = test_message()
message_ebcdic_raw = test_message('cp500')
message_ascii_raw_hex = test_message(hex_bitmap=True)
message_ebcdic_raw_hex = test_message('cp500', hex_bitmap=True)



import binascii
import datetime
import decimal
import unittest

from cardutil import CardutilError
from cardutil.config import config
from cardutil.iso8583 import (
    BitArray, _iso8583_to_field, _field_to_iso8583, _iso8583_to_dict, _dict_to_iso8583, loads, dumps,
    _pds_to_de, _pds_to_dict, _pytype_to_string, _icc_to_dict, _get_de43_fields, _get_date_from_string)

from tests import message_ebcdic_raw, message_ascii_raw, message_ascii_raw_hex, message_ebcdic_raw_hex


class Iso8583TestCase(unittest.TestCase):
    def test_dumps(self):
        # use config from package
        out_data = dumps({'MTI': '1234', 'DE2': '123'})
        self.assertEqual(b'1234\xc0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0003123', out_data)

        # explicitly pass the config
        out_data = dumps({'MTI': '1234', 'DE2': '123'}, iso_config=config["bit_config"])
        self.assertEqual(b'1234\xc0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0003123', out_data)

    def test_dumps_with_no_values(self):
        # check that empty fields are not added as elements to ISO output
        in_data_empty = {'MTI': '1234', 'DE2': '', 'DE3': None, 'DE4': 0}
        out_data_empty = loads(dumps(in_data_empty))
        print(out_data_empty)
        self.assertEqual({'MTI': '1234', 'DE4': 0}, out_data_empty)

    def test_dumps_hex_bitmap(self):
        out_data = dumps({'MTI': '1234', 'DE2': '123'}, hex_bitmap=True)
        self.assertEqual(b'1234c000000000000000000000000000000003123', out_data)

    def test_loads(self):
        self.assertEqual(
            {'MTI': '1234', 'DE2': '123'}, loads(dumps({'MTI': '1234', 'DE2': '123'}), iso_config=config['bit_config']))
        self.assertEqual({'MTI': '1234', 'DE2': '123'}, loads(dumps({'MTI': '1234', 'DE2': '123'})))

    def test_dump_load_ascii(self):
        self.assertEqual(message_ascii_raw, dumps(loads(message_ascii_raw)))
        self.assertEqual(message_ascii_raw_hex, dumps(loads(message_ascii_raw), hex_bitmap=True))

    def test_dump_load_ebcdic(self):
        self.assertEqual(message_ebcdic_raw, dumps(loads(message_ebcdic_raw, encoding='cp500'), encoding='cp500'))
        self.assertEqual(
            message_ebcdic_raw_hex,
            dumps(loads(message_ebcdic_raw_hex, encoding='cp500', hex_bitmap=True), encoding='cp500', hex_bitmap=True))

    def test_bitarray(self):
        """
        feed in binary bitmap, convert to array then back to bitmap. Make sure its the same
        """
        start_bitmap = b'\xF0\x10\x05\x42\x84\x61\x80\x02\x02\x00\x00\x04\x00\x00\x00\x00'
        test_bitarray = BitArray()
        test_bitarray.frombytes(start_bitmap)
        bit_list = test_bitarray.tolist()
        for index, value in enumerate(bit_list):
            if value:
                print(f'{index+1}={value}')
        expected_true_bits = (1, 2, 3, 4, 12, 22, 24, 26, 31, 33, 38, 42, 43, 48, 49, 63, 71, 94)
        expected_bits = [True if bit+1 in expected_true_bits else False for bit in range(128)]
        self.assertEqual(expected_bits, bit_list)
        test_bitarray.fromlist(expected_bits)
        end_bitmap = test_bitarray.tobytes()
        self.assertEqual(start_bitmap, end_bitmap)

    def test_iso8583_to_field(self):
        self.assertEqual(
            ({'DE1': '4564320012321122'}, 18),
            _iso8583_to_field('1', {'field_type': 'LLVAR', 'field_length': 0}, b'164564320012321122'))
        self.assertEqual(
            ({'DE1': '456432******1122'}, 18),
            _iso8583_to_field(
                '1', {'field_type': 'LLVAR', 'field_length': 0, 'field_processor': 'PAN'}, b'164564320012321122'))
        self.assertEqual(
            ({'DE1': '456432001'}, 18),
            _iso8583_to_field(
                '1', {'field_type': 'LLVAR', 'field_length': 0, 'field_processor': 'PAN-PREFIX'},
                b'164564320012321122'))
        self.assertEqual(
            ({'DE1': b'\x01\x01\xff', 'ICC_DATA': '0101ff', 'TAG01': 'ff'}, 5),
            _iso8583_to_field(
                '1', {'field_type': 'LLVAR', 'field_length': 0, 'field_processor': 'ICC'}, b'03\x01\x01\xFF'))
        self.assertEqual(
            ({'DE1': '4564320012321122'}, 19),
            _iso8583_to_field('1', {'field_type': 'LLLVAR', 'field_length': 0}, b'0164564320012321122'))
        self.assertEqual(
            ({'DE1': '4564320012321122    '}, 20),
            _iso8583_to_field('1', {'field_type': 'FIXED', 'field_length': 20}, b'4564320012321122    '))
        self.assertEqual(
            ({'DE1': 1234}, 20),
            _iso8583_to_field(
                '1', {'field_type': 'FIXED', 'field_python_type': 'int', 'field_length': 20}, b'00000000000000001234'))
        self.assertEqual(
            ({'DE1': 1234}, 6),
            _iso8583_to_field('1', {'field_type': 'LLVAR', 'field_python_type': 'int', 'field_length': 0}, b'041234'))
        self.assertEqual(
            ({'DE1': decimal.Decimal('123.432')}, 20),
            _iso8583_to_field('1', {'field_type': 'FIXED', 'field_python_type': 'decimal', 'field_length': 20},
                              b'0000000000000123.432'))
        # trigger decode error field length
        with self.assertRaises(CardutilError):
            _iso8583_to_field('1', {'field_type': 'LLVAR', 'field_length': 0}, b'\x00\x004564320012321122')
        # unable to decode field length
        with self.assertRaises(CardutilError):
            _iso8583_to_field('01', {"field_type": "LLLVAR", "field_length": 255}, b'01\xff', encoding='ascii')
        # invalid field length
        with self.assertRaises(CardutilError):
            _iso8583_to_field('01', {"field_type": "LLLVAR", "field_length": 255}, b'01X.', encoding='ascii')
        # unable to decode field value
        with self.assertRaises(CardutilError):
            _iso8583_to_field('01', {"field_type": "LLLVAR", "field_length": 255}, b'001\xff', encoding='ascii')
        # unable to convert to python value

        with self.assertRaises(CardutilError):
            _iso8583_to_field('1', {'field_type': 'LLVAR', 'field_python_type': 'int', 'field_length': 0}, b'04XXXX')

    def test_field_to_iso8583(self):
        self.assertEqual(b'164564320012321122', _field_to_iso8583({'field_type': 'LLVAR'}, "4564320012321122"))
        self.assertEqual(
            b'164564320012321122', _field_to_iso8583({'field_type': 'LLVAR', 'field_length': 0}, "4564320012321122"))
        self.assertEqual(b'0164564320012321122', _field_to_iso8583({'field_type': 'LLLVAR'}, "4564320012321122"))
        self.assertEqual(
            b'0164564320012321122', _field_to_iso8583({'field_type': 'LLLVAR', 'field_length': 0}, "4564320012321122"))
        self.assertEqual(
            b'4564320012321122    ', _field_to_iso8583({'field_type': 'FIXED', 'field_length': 20}, "4564320012321122"))
        self.assertEqual(
            b'00000000000000001234', _field_to_iso8583(
                {'field_type': 'FIXED', 'field_python_type': 'int', 'field_length': 20}, 1234))
        self.assertEqual(
            b'041234', _field_to_iso8583({'field_type': 'LLVAR', 'field_python_type': 'int'}, 1234))
        self.assertEqual(
            b'041234', _field_to_iso8583({'field_type': 'LLVAR', 'field_python_type': 'int'}, '1234'))
        # TODO Exception if field overflow
        self.assertEqual(
            b'123', _field_to_iso8583({'field_type': 'FIXED', 'field_python_type': 'int', 'field_length': 3}, 1234))
        self.assertEqual(
            b'01234', _field_to_iso8583({'field_type': 'FIXED', 'field_python_type': 'int', 'field_length': 5}, 1234))

        self.assertEqual(
            b'041234', _field_to_iso8583({'field_type': 'LLVAR', 'field_python_type': 'int', 'field_length': 0}, 1234))

        self.assertEqual(
            b'0000000000000123.432', _field_to_iso8583(
                {'field_type': 'FIXED', 'field_python_type': 'decimal', 'field_length': 20},
                decimal.Decimal("123.432")))
        self.assertEqual(
            b'140102151610', _field_to_iso8583(
                {'field_type': 'FIXED', 'field_python_type': 'datetime',
                 'field_length': 12, "field_date_format": "%y%m%d%H%M%S"},
                datetime.datetime(2014, 1, 2, 15, 16, 10)))

    def test_iso8583_to_dict(self):
        expected_dict = {'MTI': '1144', 'DE2': '4444555544445555', 'DE3': '111111', 'DE4': 9999,
                         'DE12': datetime.datetime(2015, 8, 15, 17, 15, 0),
                         'DE22': '123456789012', 'DE24': '333', 'DE26': 1234,
                         'DE31': '57995799120000001230612', 'DE33': '123456', 'DE38': '123456',
                         'DE42': '579942111111111', 'DE43': 'BIG BOBS\\80 KERNDALE ST\\DANERLEY\\3103  VICAUS',
                         'DE43_NAME': 'BIG BOBS', 'DE43_ADDRESS': '80 KERNDALE ST', 'DE43_SUBURB': 'DANERLEY',
                         'DE43_POSTCODE': '3103', 'DE43_STATE': 'VIC', 'DE43_COUNTRY': 'AUS',
                         'DE48': '0001001Y', 'PDS0001': 'Y', 'DE49': '999', 'DE63': '0000000000000001',
                         'DE71': 12345678, 'DE94': '999999'}

        ascii_dict = _iso8583_to_dict(message_ascii_raw, config["bit_config"], "ascii")
        self.assertEqual(ascii_dict, expected_dict)
        ebcdic_dict = _iso8583_to_dict(message_ebcdic_raw, config["bit_config"], "cp500")
        self.assertEqual(ebcdic_dict, expected_dict)

        # check exception when full message not processed
        with self.assertRaises(CardutilError):
            _iso8583_to_dict(message_ascii_raw + b' ', config["bit_config"], "ascii")

        # check exception when cannot unpack mti and bitmap (struct error)
        with self.assertRaises(CardutilError):
            _iso8583_to_dict(b'\x00TI0', config["bit_config"], "ascii")

        # check exception when cannot decode MTI
        with self.assertRaises(CardutilError):
            _iso8583_to_dict(b'\xFFBCDFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF.....', config["bit_config"],
                             encoding="ascii", hex_bitmap=True)

    def test_dict_to_iso8583(self):
        source_dict = {'MTI': '1144', 'DE2': '4444555544445555', 'DE3': '111111', 'PDS0001': '1', 'PDS9999': 'Z'}
        actual_iso = _dict_to_iso8583(source_dict, config['bit_config'])
        expected_iso = (b'1144\xe0\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00164444555544445555111111'
                        b'016000100119999001Z')
        print(actual_iso)
        self.assertEqual(expected_iso, actual_iso)

    def test_dict_to_pds_to_de(self):
        vals = {'PDS0001': '123', 'PDS9999': 'ABCDEF'}
        outs = _pds_to_de(vals)
        print(outs)
        self.assertEqual(_pds_to_dict(outs.pop()), vals)

    def test_pds_to_de_multiple_fields(self):
        vals = {'PDS0001': '*' * 900, 'PDS9999': '!' * 900}
        outs = _pds_to_de(vals)
        print(outs)
        self.assertEqual(_pds_to_dict(outs.pop()), {'PDS9999': '!' * 900})
        self.assertEqual(_pds_to_dict(outs.pop()), {'PDS0001': '*' * 900})

    def test_pds_to_de_no_pds_fields(self):
        vals = {'DE1': '*', 'DE2': '*'}
        outs = _pds_to_de(vals)
        print(outs)
        self.assertListEqual(outs, [])

    def test_pytype_to_string(self):
        self.assertEqual('ABC', _pytype_to_string('ABC', {'field_python_type': 'string', 'field_length': 5}))
        self.assertEqual('ABC', _pytype_to_string('ABC', {'field_length': 5}))
        self.assertEqual('00001', _pytype_to_string('1', {'field_python_type': 'int', 'field_length': 5}))
        self.assertEqual('00001', _pytype_to_string('1', {'field_python_type': 'long', 'field_length': 5}))
        self.assertEqual(
            '180101171500', _pytype_to_string(datetime.datetime(2018, 1, 1, 17, 15), {
                'field_python_type': 'datetime', 'field_date_format': '%y%m%d%H%M%S'}))

    def test_icc_to_dict(self):
        self.assertEqual(
            _icc_to_dict(b'\x01\x01\x41\x9f\x01\x02\x12\x34'),
            {'ICC_DATA': '0101419f01021234', 'TAG01': '41', 'TAG9F01': '1234'})

    def test_icc_to_dict_null_eof(self):
        # issue and test data provided by Diego Felipe Maia (diegofmaia28@hotmail.com)
        test_de55 = binascii.unhexlify(
            '9f26081c89a48c0c4c3a309f2701809f10120110a04001240000000000000000000000ff9f370401'
            'bd0f6d9f36020011950500000480009a031909069c01009f02060000000289985f2a020032820239'
            '009f1a0200329f03060000000000009f1e0837383636343738349f3303e0f0c89f3501009f090200'
            '009f34034203008407a0000000041010910aaf6f5977b4ca29250012000000000000000000000000'
            '00000000000000000000000000000000000000000000000000000000000000000000000000000000'
            '00000000000000000000000000000000000000000000000000000000000000000000000000000000'
            '000000000000000000000000000000')
        print(_icc_to_dict(test_de55))

    def test_get_de43_fields(self):
        # does not match
        self.assertEqual(_get_de43_fields('THIS DOES NOT MATCH REGEX'), {})
        # matches regex
        expected_dict = {'DE43_ADDRESS': '100 MAIN ST',
                         'DE43_COUNTRY': 'AUS',
                         'DE43_NAME': 'BOBS BURGERS',
                         'DE43_POSTCODE': '4102',
                         'DE43_STATE': 'QLD',
                         'DE43_SUBURB': 'WOOLLOONGABBA'}
        self.assertEqual(_get_de43_fields('BOBS BURGERS\\100 MAIN ST\\WOOLLOONGABBA\\4102 QLDAUS'), expected_dict)
        processor_config = r'(?P<DE43_ALL>.*)'
        self.assertEqual(_get_de43_fields('ALL FIELD', processor_config=processor_config), {'DE43_ALL': 'ALL FIELD'})

    def test_get_date_from_string_use_fromisodate(self):
        import builtins
        import sys

        # skip test if being run on py36 or less
        if sys.version_info < (3, 7):
            self.skipTest('Needs py37 or greater to run')

        real_import = builtins.__import__

        def mock_import(name, *vars):
            if name == 'dateutil.parser':
                raise ImportError
            else:
                return real_import(name, *vars)

        builtins.__import__ = mock_import
        self.get_date_from_string()
        builtins.__import__ = real_import

    def test_get_date_from_string_use_builtin(self):
        import sys
        import builtins

        real_version = sys.version_info
        sys.version_info = (3, 6)
        real_import = builtins.__import__

        def mock_import(name, *vars):
            if name == 'dateutil.parser':
                raise ImportError
            else:
                return real_import(name, *vars)

        builtins.__import__ = mock_import
        self.get_date_from_string()
        with self.assertRaises(ValueError):
            _get_date_from_string("11221-11-22")
        builtins.__import__ = real_import
        sys.version_info = real_version

    def test_get_date_from_string_use_dateutil(self):
        try:
            import dateutil
        except ImportError:
            self.skipTest('requires that python-dateutil is installed')
            return
        print(dateutil)
        self.get_date_from_string()

    def get_date_from_string(self):
        self.assertEqual(_get_date_from_string("2002-01-01"), datetime.datetime(2002, 1, 1))
        self.assertEqual(_get_date_from_string("2002-01-01 10:01"), datetime.datetime(2002, 1, 1, 10, 1))
        self.assertEqual(_get_date_from_string("2002-01-01 10:01:02"), datetime.datetime(2002, 1, 1, 10, 1, 2))


if __name__ == '__main__':
    unittest.main()



import unittest

from cardutil import card


class CardTestCase(unittest.TestCase):
    def test_check_digit(self):
        acct = '7992739871'
        self.assertEqual('3', card.calculate_check_digit(acct))
        self.assertIsNone(card.validate_check_digit(card.add_check_digit(acct)))
        with self.assertRaises(AssertionError):
            card.validate_check_digit("111")

    def test_mask(self):
        self.assertEqual('123456**9012', card.mask('123456789012'))
        self.assertEqual('123456..9012', card.mask('123456789012', mask_char='.'))


if __name__ == '__main__':
    unittest.main()



import io
import os
import tempfile
import unittest

from cardutil import CardutilError
from cardutil.mciipm import VbsWriter
from cardutil.cli import mci_ipm_param_to_csv
from cardutil.config import config

MCI_PARAMETER_CONFIG = config.get('mci_parameter_tables')


class MciIpmParamToCsvTestCase(unittest.TestCase):
    def test_mci_ipm_param_to_csv_parser(self):
        args = vars(mci_ipm_param_to_csv.cli_parser().parse_args(['file1.ipm', 'IP0000T1']))
        self.assertEqual(
            args,
            {'in_filename': 'file1.ipm', 'out_filename': None, 'in_encoding': None, 'table_id': 'IP0000T1',
             'out_encoding': None, 'no1014blocking': False, 'config_file': None, 'debug': False})

    def test_ipm_to_csv_input_params(self):
        """
        Actually run using real files
        :return:
        """
        param_file_data = (
            b'2011101414AIP0000T1IP0000T1 TABLE LIST                 ' + 188 * b'.' + b'001',
            b'2014101414AIP0000T1IP0040T1 ACCOUNT RANGE TABLE        ' + 188 * b'.' + b'036',
            b'TRAILER RECORD IP0000T1  00000218                                               ',
            b'........xxx....',  # dummy record
            b'1711114A0365116545113000000000MCC5116545113999999999MCC020000000152710084563AUS036CMCC NNYMCC N0000000362'
            b'0000000000000000000000000000 000000NN   000000NNNN0NUNN0N N     ',
        )

        with tempfile.NamedTemporaryFile(mode='w+b', delete=False) as in_file:
            with VbsWriter(in_file, blocked=True) as in_file_writer:
                in_file_writer.write_many(param_file_data)
            in_file_name = in_file.name
            print(in_file_name)
            in_file.close()
        mci_ipm_param_to_csv.cli_run(table_id='IP0040T1', in_filename=in_file_name)
        mci_ipm_param_to_csv.cli_run(table_id='IP0040T1', in_filename=in_file_name, in_encoding='latin_1')
        mci_ipm_param_to_csv.cli_run(table_id='IP0040T1', in_filename=in_file_name, out_filename=in_file_name + 'xxx')
        mci_ipm_param_to_csv.cli_run(table_id='IP0040T1', in_filename=in_file_name, debug=True)
        os.remove(in_file_name)
        os.remove(in_file_name + '_IP0040T1.csv')
        os.remove(in_file_name + 'xxx')

    def test_extract_ip0040t1(self):
        param_file_data = [
            b'2011101414AIP0000T1IP0000T1 TABLE LIST                 ' + 188 * b'.' + b'001',
            b'2014101414AIP0000T1IP0040T1 ACCOUNT RANGE TABLE        ' + 188 * b'.' + b'036',
            b'TRAILER RECORD IP0000T1  00000218                                               ',
            b'........xxx....',  # dummy record
            b'1711114A0365116545113000000000MCC5116545113999999999MCC020000000152710084563AUS036CMCC NNYMCC N0000000362'
            b'0000000000000000000000000001 000000NY   000000NNNN',
            b'0000000000000000000000000000 000000NN   000000NNNN0NUNN0N N     ',
        ]

        with io.BytesIO() as test_param_stream, io.StringIO() as test_csv_stream:
            with VbsWriter(test_param_stream, blocked=True) as test_param_vbs:
                test_param_vbs.write_many(param_file_data)

            test_param_stream.seek(0)
            mci_ipm_param_to_csv.mci_ipm_param_to_csv(
                test_param_stream, test_csv_stream, config=MCI_PARAMETER_CONFIG, table_id='IP0040T1')
            test_csv_stream.seek(0)
            test_csv_data = test_csv_stream.read()

        csv_records = test_csv_data.split('\n')

        # check the header
        config_header_keys = list(MCI_PARAMETER_CONFIG.get('IP0040T1').keys())
        csv_header_keys = csv_records[0].split(',')
        self.assertEqual(config_header_keys, csv_header_keys)

        # check the record extracted
        self.assertEqual(csv_records[1], '711114,A,036,5116545113000000000,MCC,5116545113999999999,MCC,02,00000001527,'
                                         '1,0084563,AUS,036,C,MCC, ,N,N,Y,MCC, ,N,000000,036,2,'
                                         '0000000000000000000000000001, ,000000,N,Y,   ,000000,N,N,N,N')

    def test_extract_ip0040t1_no_records(self):
        param_file_data = [
            b'2011101414AIP0000T1IP0000T1 TABLE LIST                 ' + 188 * b'.' + b'001',
            b'2014101414AIP0000T1IP0040T1 ACCOUNT RANGE TABLE        ' + 188 * b'.' + b'036',
            b'TRAILER RECORD IP0000T1  00000218                                               ',
        ]

        with io.BytesIO() as test_param_stream, io.StringIO() as test_csv_stream:
            with VbsWriter(test_param_stream, blocked=True) as test_param_vbs:
                test_param_vbs.write_many(param_file_data)

            test_param_stream.seek(0)
            mci_ipm_param_to_csv.mci_ipm_param_to_csv(
                test_param_stream, test_csv_stream, config=MCI_PARAMETER_CONFIG, table_id='IP0040T1')
            test_csv_stream.seek(0)
            test_csv_data = test_csv_stream.read()

        csv_records = test_csv_data.split('\n')
        print(csv_records)
        self.assertEqual(len(csv_records), 2, msg="Only 2 record returned -- header, dummy")

        # check the header
        config_header_keys = list(MCI_PARAMETER_CONFIG.get('IP0040T1').keys())
        csv_header_keys = csv_records[0].split(',')
        self.assertEqual(config_header_keys, csv_header_keys)

        # check the dummy value
        self.assertEqual(csv_records[1], '')

    def test_extract_ip0040t1_no_table_config(self):
        with io.BytesIO() as test_param_stream, io.StringIO() as test_csv_stream:
            with self.assertRaises(CardutilError):
                mci_ipm_param_to_csv.mci_ipm_param_to_csv(
                    test_param_stream, test_csv_stream, config=MCI_PARAMETER_CONFIG, table_id='BADTABLE')

    def test_extract_ip0040t1_missing_index_trailer(self):
        param_file_data = [
            b'2011101414AIP0000T1IP0000T1 TABLE LIST                 ' + 188 * b'.' + b'001',
            b'2014101414AIP0000T1IP0040T1 ACCOUNT RANGE TABLE        ' + 188 * b'.' + b'036']

        with io.BytesIO() as test_param_stream, io.StringIO() as test_csv_stream:
            with VbsWriter(test_param_stream, blocked=True) as test_param_vbs:
                test_param_vbs.write_many(param_file_data)

            test_param_stream.seek(0)
            with self.assertRaises(CardutilError):
                mci_ipm_param_to_csv.mci_ipm_param_to_csv(
                    test_param_stream, test_csv_stream, config=MCI_PARAMETER_CONFIG, table_id='IP0040T1')


if __name__ == '__main__':
    unittest.main()



import argparse
import collections
import csv
import logging

from cardutil import __version__
from cardutil.cli import get_config, print_banner, print_exception_details
from cardutil.mciipm import IpmReader, IpmWriter, MciIpmDataError


def cli_entry(*args):
    return cli_run(**vars(_get_cli_parser().parse_args(*args)))


def cli_run(**kwargs):

    print_banner('mideu', kwargs)
    if not len(kwargs):
        print(__name__ + " (" + __version__ + ")")
        print("try --help for information")
        return -1

    config = get_config('cardutil.json', cli_filename=kwargs.get('config_file'))

    logging.basicConfig(
        level=kwargs.get('loglevel'),
        format='%(asctime)s:%(name)s:%(lineno)s:%(levelname)s:%(message)s')

    try:
        kwargs['func'](config=config, **kwargs)
    except MciIpmDataError as err:
        print_exception_details(err)
        return -1


def dicts_to_csv(data_list, output_file, field_list=None):
    """
    Writes dict data to CSV file

    :param data_list: list of dictionaries that contain the data to be loaded
    :param output_file: output CSV file
    :param field_list: (optional) list of fields to output to CSV file
    :return: None
    """

    if not field_list:
        # get the fields present in the dicts
        field_summary = collections.Counter()
        data_list = list(data_list)
        for item in data_list:
            field_summary.update(item.keys())
        field_list = [field_key for field_key in field_summary]

    writer = csv.DictWriter(
        output_file,
        fieldnames=field_list,
        extrasaction="ignore",
        lineterminator="\n")

    writer.writeheader()
    for data_item in data_list:
        writer.writerow({item: data_item[item] for item in data_item if item in field_list})


def _get_cli_parser():
    """
    mideu argparse parser create
    :return: parser
    """
    parser = argparse.ArgumentParser(
        description="MasterCard IPM file formatter ({version})".format(
            version=__version__)
    )
    parser.add_argument("--version", action="version",
                        version="%(prog)s (" + __version__ + ")",
                        help="Get version information")

    subparsers = parser.add_subparsers(help="Sub-command help")

    # Extract command
    extract_parser = subparsers.add_parser("extract", help="Extract help")
    extract_parser.set_defaults(func=extract)
    _add_common_args(extract_parser)
    _add_extract_args(extract_parser)

    # Convert command
    convert_parser = subparsers.add_parser("convert", help="Convert help")
    convert_parser.set_defaults(func=convert)
    _add_common_args(convert_parser)

    return parser


def _add_common_args(parser):
    """
    mideu add common cli arguments
    :param parser: the argparse parser
    :return: None
    """
    parser.add_argument("input", help="Input IPM file name")
    add_source_format_arg(parser)
    add_logging_arg_group(parser)


def add_source_format_arg(parser):
    """
    Adds source format option to parser
    :param parser: the parser to add the source format option to
    :return: None
    """
    parser.add_argument(
        "-s", "--sourceformat",
        help="encoding format of source file",
        choices=["ebcdic", "ascii"],
        default="ebcdic"
    )

    parser.add_argument(
        "--no1014blocking",
        help="do not use 1014 block format. Just vbs type record",
        dest="no1014blocking",
        action="store_true"
    )


def add_logging_arg_group(parser):
    """
    Adds logging options to parser
    :param parser: the argparse parser to add logging options to
    :return: None
    """
    logging_arg_group = parser.add_argument_group("logging options")
    logging_arg_group.add_argument("-d", "--debug",
                                   help="turn debugging output on",
                                   action="store_const",
                                   dest="loglevel",
                                   const=logging.DEBUG,
                                   default=logging.WARNING)
    logging_arg_group.add_argument("-v", "--verbose",
                                   help="turn information output on",
                                   action="store_const",
                                   dest="loglevel",
                                   const=logging.INFO)


def _add_extract_args(parser):
    """
    mideu add extract subcommand arguments
    :param parser: the argparse parser
    :return: None
    """
    csv_arg_group = parser.add_argument_group("csv output options")
    csv_arg_group.add_argument("--csvoutputfile", help="Output filename")


def extract(config, **kwargs):
    """
    Create a csv file given an input Mastercard IPM file

    :param config: dict containing cardutil config
    :return: None
    """
    blocked = not kwargs.get('no1014blocking', False)
    kwargs['out_encoding'] = 'utf8'
    if kwargs.get('sourceformat', 'ebcdic') == 'ebcdic':
        kwargs['in_encoding'] = 'cp500'
    else:
        kwargs['in_encoding'] = 'latin1'
    if not kwargs.get('csvoutputfile'):
        kwargs['csvoutputfile'] = kwargs['input'] + '.csv'

    with open(kwargs['input'], 'rb') as in_ipm:
        with open(kwargs['csvoutputfile'], 'w', encoding=kwargs.get('out_encoding')) as out_csv:
            dicts_to_csv(
                IpmReader(in_ipm, encoding=kwargs['in_encoding'], blocked=blocked, iso_config=config.get('bit_config')),
                out_csv, field_list=config.get('output_data_elements'))


def convert(config, **kwargs):
    """
    Change encoding of IPM file from one encoding scheme to another

    :return: None
    """
    in_filename = kwargs['input']
    out_filename = in_filename + '.out'
    in_blocked = not kwargs.get('no1014blocking', False)
    out_blocked = in_blocked
    if kwargs.get('sourceformat', 'ebcdic') == 'ebcdic':
        in_encoding = 'cp500'
        out_encoding = 'latin1'
    else:
        in_encoding = 'latin1'
        out_encoding = 'cp500'

    with open(in_filename, 'rb') as in_file:
        with open(out_filename, 'wb') as out_file:
            with IpmWriter(out_file, encoding=out_encoding, blocked=in_blocked) as writer:
                reader = IpmReader(in_file, encoding=in_encoding, blocked=out_blocked)
                writer.write_many(reader)



"""
Provides services related to card numbers.

* check digits
* masking

"""
from itertools import cycle


def calculate_check_digit(card_number: str) -> str:
    """
    calculate luhn 10 check digit.

    :param card_number: number to calculate check digit for excluding check digit.
    :return: check digit value
    """
    digits = [int(digit) for digit in card_number if digit.isdigit()]  # change string to list of integers
    total = sum([sum(divmod(multiplier * digit, 10)) for digit, multiplier in zip(digits[::-1], cycle([2, 1]))])
    return str((total * 9) % 10)


def validate_check_digit(card_number: str) -> str:
    """
    validate luhn 10 check digit

    :param card_number: number with check digit
    :return: None
    :raises AssertionError: Check digit is not valid
    """
    assert calculate_check_digit(card_number[0:-1]) == card_number[-1]


def add_check_digit(card_number: str) -> str:
    """
    adds luhn 10 check digit

    :param card_number: number to add check digit for
    :return: number with check digit
    """
    return card_number + calculate_check_digit(card_number)


def mask(card_number: str, mask_char: str = '*') -> str:
    """
    returns a masked version of a card number
    Format is First 6, last 4 digits.

    :param card_number: the card number to mask
    :param mask_char: the character to use in masking
    :return: masked card number
    """
    return card_number[0:6] + mask_char * (len(card_number)-10) + card_number[-4:]



"""
The config module contains the default configuration for the cardutil library.

The config consists of a single dictionary called `config` with a number of keys for different configuration data.

bit_config
==========

The iso8583 module functions require configuration that details how to process the bitmap in a message.

Field definitions in an ISO8583 message are generally the same but there can be variations you may need to cater for
with the different card schemes. The library supports provision of custom ISO8583 configuration.

If a configuration is not provided, the variable ``config['bit_config']`` in ``config.py`` provides a Mastercard
specific ISO8583 configuration.

The config is in the form of a python dictionary structured as follows

.. code-block:: json

    {
        "1": {
            "field_name": "Bitmap secondary",
            "field_type": "FIXED",
            "field_length": 8},
        "2": {
            "field_name": "PAN",
            "field_type": "LLVAR",
            "field_length": 0,
            "field_processor": "PAN",
            "field_processor_config": "",
            "field_python_type": "string",
            "field_date_format": "%y%m%d"}
    }

The config dictionary contains a string key for each valid bit in the bitmap.
For example, the config for bit 1 with have the key "1".

The value for the bit is a dictionary containing the following values:

field_name
    description of the bit field
field_type
    defines the ISO8583 field type
        * FIXED : field with fixed length
        * LLVAR : field with variable length - 2 character length value
        * LLLVAR : field with variable length - 2 character length value
field_length
    length of the field. Use zero for variable length fields.
field_processor
    (optional) a process to apply to a field.
        * ``PAN``: For use with PAN fields. Mask PAN using first 6, last 4 pattern
        * ``PAN-PREFIX``: For use with PAN fields. Only get first 9 PAN numbers -- prefix
        * ``ICC``: Mastercard ICC field. Adds TAGxxxx keys to output
        * ``PDS``: Mastercard PDS field. Processes PDSxxxx fields
        * ``DE43``: Mastercard Merchant details field. Adds DE43 sub fields DE43_*
field_processor_config
     (optional) Where a field processor requires extra config, it can be placed in this field.
     For DE43 processor:
        field should contain a regex to split the DE43 field up into components. The regex
        groups defined will be added to the returned dictionary. Use Python regex variation.

field_python_type
    (optional) When processing between iso and python, determines the python object type.
        * ``string``: default if no type provided
        * ``int``
        * ``decimal``: use if your field has decimal place
        * ``datetime``: use if your field is a date
field_date_format
     (optional) If your field python type is datetime, then you use this to specify
     the format of the date in the iso record.

     Use standard python datetime.strptime see `strftime() and strptime() Behaviour
     <https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior>`_.

     Default format is "%y%m%d"

output_data_elements
====================

This config defines the ISO elements that will be present in CSV file output. Is a list of the field keys

.. code-block:: json

    ["MTI", "DE2", "DE3", "DE4", "DE12", "DE14", "DE22", "DE23", "DE24", "DE25", "DE26"]


mci_parameter_tables
====================

Provides configuration required to extract IPM parameter extracts

.. code-block:: json

    {
        "IP0006T1": {
            "effective_timestamp": {"start": 1, "end": 10},
            "active_inactive_code": {"start": 7, "end": 8},
            "table_id": {"start": 8, "end": 11},
            "card_program_id": {"start": 11, "end": 14},
            "data_element_id": {"start": 14, "end": 17},
            "data_element_name": {"start": 17, "end": 74},
            "data_element_format": {"start": 74, "end": 77}
        },
        "IP0040T1": {}
    }

"""

config = {
    "bit_config": {
        "1": {"field_name": "Bitmap secondary", "field_type": "FIXED", "field_length": 8},
        "2": {"field_name": "PAN", "field_type": "LLVAR", "field_length": 0},  # "field_processor": "PAN"},
        "3": {"field_name": "Processing code", "field_type": "FIXED", "field_length": 6},
        "4": {"field_name": "Amount transaction", "field_type": "FIXED", "field_length": 12,
              "field_python_type": "long"},
        "5": {"field_name": "Amount, Reconciliation", "field_type": "FIXED", "field_length": 12,
              "field_python_type": "long"},
        "6": {"field_name": "Amount, Cardholder billing", "field_type": "FIXED", "field_length": 12,
              "field_python_type": "long"},
        "9": {"field_name": "Conversion rate, Reconciliation", "field_type": "FIXED", "field_length": 8,
              "field_python_type": "long"},
        "10": {"field_name": "Conversion rate, Cardholder billing", "field_type": "FIXED", "field_length": 8,
               "field_python_type": "long"},
        "12": {"field_name": "Date/Time local transaction", "field_type": "FIXED", "field_length": 12,
               "field_python_type": "datetime", "field_date_format": "%y%m%d%H%M%S"},
        "14": {"field_name": "Expiration date", "field_type": "FIXED", "field_length": 4},
        "22": {"field_name": "Point of service data code", "field_type": "FIXED", "field_length": 12},
        "23": {"field_name": "Card sequence number", "field_type": "FIXED", "field_length": 3},
        "24": {"field_name": "Function code", "field_type": "FIXED", "field_length": 3},
        "25": {"field_name": "Message reason code", "field_type": "FIXED", "field_length": 4},
        "26": {"field_name": "Card acceptor business code", "field_type": "FIXED", "field_length": 4,
               "field_python_type": "int"},
        "30": {"field_name": "Amounts, original", "field_type": "FIXED", "field_length": 24},
        "31": {"field_name": "Acquirer reference data", "field_type": "LLVAR", "field_length": 23},
        "32": {"field_name": "Acquiring institution ID code", "field_type": "LLVAR", "field_length": 0},
        "33": {"field_name": "Forwarding institution ID code", "field_type": "LLVAR", "field_length": 0},
        "37": {"field_name": "Retrieval reference number", "field_type": "FIXED", "field_length": 12},
        "38": {"field_name": "Approval code", "field_type": "FIXED", "field_length": 6},
        "40": {"field_name": "Service code", "field_type": "FIXED", "field_length": 3},
        "41": {"field_name": "Card acceptor terminal ID", "field_type": "FIXED", "field_length": 8},
        "42": {"field_name": "Card acceptor Id", "field_type": "FIXED", "field_length": 15},
        "43": {"field_name": "Card acceptor name/location", "field_type": "LLVAR", "field_length": 0,
               "field_processor": "DE43",
               "field_processor_config": r"(?P<DE43_NAME>.+?) *\\(?P<DE43_ADDRESS>.+?) *\\(?P<DE43_SUBURB>.+?) *\\"
                                         r"(?P<DE43_POSTCODE>\S{4,10}) *(?P<DE43_STATE>.{3})(?P<DE43_COUNTRY>.{3})"},
        "48": {"field_name": "Additional data", "field_type": "LLLVAR", "field_length": 0, "field_processor": "PDS"},
        "49": {"field_name": "Currency code, Transaction", "field_type": "FIXED", "field_length": 3},
        "50": {"field_name": "Currency code, Reconciliation", "field_type": "FIXED", "field_length": 3},
        "51": {"field_name": "Currency code, Cardholder billing", "field_type": "FIXED", "field_length": 3},
        "54": {"field_name": "Amounts, additional", "field_type": "LLLVAR", "field_length": 0},
        "55": {"field_name": "ICC system related data", "field_type": "LLLVAR", "field_length": 255,
               "field_processor": "ICC"},
        "62": {"field_name": "Additional data 2", "field_type": "LLLVAR", "field_length": 0, "field_processor": "PDS"},
        "63": {"field_name": "Transaction lifecycle Id", "field_type": "LLLVAR", "field_length": 16},
        "71": {"field_name": "Message number", "field_type": "FIXED", "field_length": 8, "field_python_type": "int"},
        "72": {"field_name": "Data record", "field_type": "LLLVAR", "field_length": 0},
        "73": {"field_name": "Date, Action", "field_type": "FIXED", "field_length": 6},
        "93": {"field_name": "Transaction destination institution ID", "field_type": "LLVAR", "field_length": 0},
        "94": {"field_name": "Transaction originator institution ID", "field_type": "LLVAR", "field_length": 0},
        "95": {"field_name": "Card issuer reference data", "field_type": "LLVAR", "field_length": 10},
        "100": {"field_name": "Receiving institution ID", "field_type": "LLVAR", "field_length": 11},
        "111": {"field_name": "Amount, currency conversion assignment", "field_type": "LLLVAR", "field_length": 0},
        "123": {"field_name": "Additional data 3", "field_type": "LLLVAR", "field_length": 0, "field_processor": "PDS"},
        "124": {"field_name": "Additional data 4", "field_type": "LLLVAR", "field_length": 0, "field_processor": "PDS"},
        "125": {"field_name": "Additional data 5", "field_type": "LLLVAR", "field_length": 0, "field_processor": "PDS"},
        "127": {"field_name": "Network data", "field_type": "LLLVAR", "field_length": 0}},

    "output_data_elements": [
        "MTI", "DE2", "DE3", "DE4", "DE12", "DE14", "DE22", "DE23", "DE24", "DE25", "DE26",
        "DE30", "DE31", "DE33", "DE37", "DE38", "DE40", "DE41", "DE42", "DE48", "DE49",
        "DE50", "DE63", "DE71", "DE73", "DE93", "DE94", "DE95", "DE100", "PDS0023",
        "PDS0052", "PDS0122", "PDS0148", "PDS0158", "PDS0165", "DE43_NAME", "DE43_SUBURB",
        "DE43_POSTCODE", "ICC_DATA"
    ],
    "mci_parameter_tables": {
        "IP0006T1": {
            "effective_timestamp": {"start": 1, "end": 10},
            "active_inactive_code": {"start": 7, "end": 8},
            "table_id": {"start": 8, "end": 11},
            "card_program_id": {"start": 11, "end": 14},
            "data_element_id": {"start": 14, "end": 17},
            "data_element_name": {"start": 17, "end": 74},
            "data_element_format": {"start": 74, "end": 77}
        },
        "IP0040T1": {
            "effective_timestamp": {"start": 1, "end": 7},
            "active_inactive_code": {"start": 7, "end": 8},
            "table_id": {"start": 8, "end": 11},
            "low_range": {"start": 11, "end": 30},
            "gcms_product": {"start": 30, "end": 33},
            "high_range": {"start": 33, "end": 52},
            "card_program_identifier": {"start": 52, "end": 55},
            "card_program_priority": {"start": 55, "end": 57},
            "member_id": {"start": 57, "end": 68},
            "product_type": {"start": 68, "end": 69},
            "endpoint": {"start": 69, "end": 76},
            "card_country_alpha": {"start": 76, "end": 79},
            "card_country_numeric": {"start": 79, "end": 82},
            "card_region": {"start": 82, "end": 83},
            "product_class": {"start": 83, "end": 86},
            "tran_routing_ind": {"start": 86, "end": 87},
            "first_present_reassign_ind": {"start": 87, "end": 88},
            "product_reassign_switch": {"start": 88, "end": 89},
            "pwcb_optin_switch": {"start": 89, "end": 90},
            "licenced_product_id": {"start": 90, "end": 93},
            "mapping_service_ind": {"start": 93, "end": 94},
            "alm_participation_ind": {"start": 94, "end": 95},
            "alm_activation_date": {"start": 95, "end": 101},
            "cardholder_billing_currency_default": {"start": 101, "end": 104},
            "cardholder_billing_currency_default_exponent": {"start": 104, "end": 105},
            "cardholder_bill_primary_currency": {"start": 105, "end": 133},
            "chip_to_magstripe_conversion_service_indicator": {"start": 133, "end": 134},
            "floor_exp_date": {"start": 134, "end": 140},
            "co_brand_participation_switch": {"start": 140, "end": 141},
            "spend_control_switch": {"start": 141, "end": 142},
            "merchant_cleansing_service_participation": {"start": 142, "end": 145},
            "merchant_cleansing_activation_date": {"start": 145, "end": 151},
            "paypass_enabled_indicator": {"start": 151, "end": 152},
            "rate_type_indicator": {"start": 152, "end": 153},
            "psn_route_indicator": {"start": 153, "end": 154},
            "cash_back_wo_purchase_ind": {"start": 154, "end": 155}
        },
        "IP0075T1": {
            "mcc_code": {"start": 12, "end": 16},
            "cab_code": {"start": 16, "end": 20}
        }
    }
}



import contextlib
import io
import os
import tempfile
import unittest

from cardutil.cli import mci_ipm_to_csv
from cardutil.config import config

CONFIG_DATA = """
{
    "bit_config": {
        "38": {
            "field_name": "Approval code",
            "field_type": "FIXED",
            "field_length": 6
        }
    }
}
"""


class MciIpmToCsvTestCase(unittest.TestCase):
    def test_mci_ipm_to_csv_cli_parser(self):
        args = vars(mci_ipm_to_csv.cli_parser().parse_args(['file1.ipm']))
        self.assertEqual(
            args,
            {'in_encoding': None, 'out_encoding': None, 'in_filename': 'file1.ipm', 'no1014blocking': False,
             'out_filename': None, 'config_file': None, 'debug': False})

        args = vars(mci_ipm_to_csv.cli_parser().parse_args(['file1.ipm', '--in-encoding', 'latin_1']))
        self.assertEqual(
            args,
            {'in_encoding': 'latin_1', 'out_encoding': None, 'in_filename': 'file1.ipm', 'no1014blocking': False,
             'out_filename': None, 'config_file': None, 'debug': False})

    def test_ipm_to_csv_bad_data_de38(self):
        """
        prod issue - auth code DE38 contains non-ascii characters.
        Added new out-encoding parameter to support this -- changed to latin_1 which supports full 256 bits.
        Also need to set in-encoding to latin_1 for the same reason
        I have only seen this issue on one transaction once after processing 1000's of input files
        """
        in_ipm = io.BytesIO(b'\x00\x00\x00\x1a0100\x80\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                            b'n\x9cm\x9cl\x9c\x00\x00\x00\x00')
        with tempfile.TemporaryFile(mode='w', encoding='latin_1') as out_csv:
            mci_ipm_to_csv.mci_ipm_to_csv(
                in_ipm=in_ipm, out_csv=out_csv, config=config, no1014blocking=True)

    def test_ipm_to_csv_input_params(self):
        """
        Actually run using real files
        :return:
        """
        in_ipm_data = (b'\x00\x00\x00\x1a0100\x80\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                       b'nXmXlX\x00\x00\x00\x00')

        with tempfile.NamedTemporaryFile(mode='w+b', delete=False) as in_ipm:
            in_ipm.write(in_ipm_data)
            in_ipm_name = in_ipm.name
            print(in_ipm_name)
            in_ipm.close()
        mci_ipm_to_csv.cli_run(in_filename=in_ipm_name, out_encoding='ascii')
        mci_ipm_to_csv.cli_run(in_filename=in_ipm_name, out_filename=in_ipm_name + '.csv', out_encoding='latin_1')

        # run with config file
        with tempfile.NamedTemporaryFile(mode='w+', delete=False) as in_config:
            config_filename = in_config.name
            in_config.write(CONFIG_DATA)
            in_config.close()
            mci_ipm_to_csv.cli_run(
                in_filename=in_ipm_name,
                out_filename=in_ipm_name + '.csv',
                config_file=in_config.name,
                out_encoding='latin_1',
                debug=True
            )
            os.remove(config_filename)
        with open(in_ipm_name + '.csv', 'r') as csv_data:
            csv_output = csv_data.read()
        self.assertEqual(csv_output, "MTI,DE38\n0100,nXmXlX\n")

        os.remove(in_ipm_name)
        os.remove(in_ipm_name + '.csv')

    def test_ipm_to_csv_generate_exception(self):
        """
        Actually run using real files, and exception generated
        Triggered through negative RDW on second record -- invalid record length
        :return:
        """
        in_ipm_data = (b'\x00\x00\x00\x1a0100\x80\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                       b'nXmXlX\xFF\xFF\x00\x00')

        with tempfile.NamedTemporaryFile(mode='w+b', delete=False) as in_ipm:
            in_ipm.write(in_ipm_data)
            in_ipm_name = in_ipm.name
            print(in_ipm_name)
            in_ipm.close()

        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            mci_ipm_to_csv.cli_run(in_filename=in_ipm_name, out_encoding='ascii')
            output = f.getvalue().splitlines()
        os.remove(in_ipm_name)
        os.remove(in_ipm_name + '.csv')
        print(output)
        assert len(output) == 9
        assert output[4] == '*** ERROR - processing has stopped ***'


if __name__ == '__main__':
    unittest.main()



import io
import os
import tempfile
import unittest

from cardutil.cli import mci_ipm_param_encode
from cardutil.mciipm import VbsWriter
from cardutil.vendor.hexdump import hexdump
from tests import print_stream


class MciIpmParamEncodeTestCase(unittest.TestCase):

    def test_mci_ipm_param_encode_input_params_blocked(self):
        """
        Run mci_ipm_param_encode using real files
        """
        # create an ipm file
        with tempfile.NamedTemporaryFile(mode='wb') as in_vbs:
            with VbsWriter(in_vbs, blocked=True) as writer:
                writer.write(b"Parameter message data")
            in_vbs.close()

    def test_mci_ipm_param_encode(self):
        # create test param file
        message_list = [b"Parameter message data" for _ in range(5)]
        vbs_in = io.BytesIO()
        with VbsWriter(vbs_in, blocked=True) as writer:
            for message in message_list:
                writer.write(message)

        # process the param encoding
        param_out = io.BytesIO()
        mci_ipm_param_encode.mci_ipm_param_encode(vbs_in, param_out, in_encoding='ascii', out_encoding='ascii')

        print_stream(vbs_in, "Input")
        vbs_in_value = vbs_in.read()

        print_stream(param_out, "Output")
        param_out_value = param_out.read()

        self.assertEqual(vbs_in_value, param_out_value)

    def test_mci_ipm_param_encode_parser(self):
        args = vars(mci_ipm_param_encode.cli_parser().parse_args(['file1.ipm']))
        self.assertEqual(
            args,
            {'in_filename': 'file1.ipm', 'out_filename': None, 'in_encoding': None,
             'out_encoding': None, 'no1014blocking': False,
             'in_format': '1014', 'out_format': '1014', 'debug': False})

    def test_mci_ipm_param_encode_input_params_variations(self):
        """
        Run mci_ipm_param_encode using real files
        """
        # create an ipm file
        with tempfile.NamedTemporaryFile(mode='wb', delete=False) as in_vbs:
            with VbsWriter(in_vbs, blocked=True) as writer:
                writer.write(b"Parameter message data")
            in_vbs_name = in_vbs.name
            in_vbs.close()

        output_file_name = in_vbs_name + '.out'

        # run with just input filename
        mci_ipm_param_encode.cli_run(in_filename=in_vbs_name, debug=True)
        with open(in_vbs_name + '.out', 'rb') as output_file:
            hexdump(output_file.read())
            print('*'*20)
        os.remove(output_file_name)

        # with input and output set to vbs (--no1014blocking: input is 1014, but doesn't seem to care)
        mci_ipm_param_encode.cli_run(in_filename=in_vbs_name, out_filename=output_file_name, no1014blocking=True)
        with open(in_vbs_name + '.out', 'rb') as output_file:
            hexdump(output_file.read())
            print('*'*20)
        os.remove(output_file_name)

        # with 1014 input and vbs output
        mci_ipm_param_encode.cli_run(
            in_filename=in_vbs_name, out_filename=output_file_name, in_format='1014', out_format='vbs')
        with open(in_vbs_name + '.out', 'rb') as output_file:
            hexdump(output_file.read())
            print('*'*20)
        os.remove(output_file_name)

        # with cp500 encoding (ebcdic)
        mci_ipm_param_encode.cli_run(in_filename=in_vbs_name, out_filename=output_file_name, out_encoding='cp500')
        with open(in_vbs_name + '.out', 'rb') as output_file:
            hexdump(output_file.read())
            print('*'*20)
        os.remove(output_file_name)

        os.remove(in_vbs_name)


if __name__ == '__main__':
    unittest.main()



"""
Pinblock
--------

There are many different pin block formats in use in the payment card industry.
The pinblock module provides functions for working with the various pin blocks formats in a consistent way.

Available pin block formats:

* :py:mod:`cardutil.pinblock.Iso0PinBlock`
* :py:mod:`cardutil.pinblock.Iso4PinBlock`

How to use::

    # create the pinblock instance. Inputs will vary depending on the format
    >>> pb = Iso0PinBlock(pin='1234', card_number='1111222233334444')
    >>> pb.pin
    '1234'

    # output pinblock bytes
    >>> binascii.hexlify(pb.to_bytes())
    b'041226dddccccbbb'

    # create pinblock instance from bytes
    >>> pb2 = Iso0PinBlock.from_bytes(pb.to_bytes(), card_number='1111222233334444')
    >>> pb2.pin
    '1234'


Pinblock mix-ins
----------------

Common operations associated with pin blocks include encryption and calculation of a pin verification value.

This module allows you to add pinblock encryption and pin verification calculators to a pinblock class through
the use of mixins.

Creating pinblock objects
^^^^^^^^^^^^^^^^^^^^^^^^^

How to create pinblock class with encryption and pin verification support::

    # use a predefined pinblock object
    >>> Pinblock = Iso0TDESPinBlockWithVisaPVV

    # or create your own class including required mix-ins
    >>> class PinBlock(Iso0PinBlock, TdesEncryptedPinBlockMixin, VisaPVVPinBlockMixin):
    ...    pass

    # or use the type builtin to create the class
    >>> PinBlock = type('MyPinBlock', (Iso0PinBlock, TdesEncryptedPinBlockMixin, VisaPVVPinBlockMixin), {})

    # create the pinblock instance. Inputs will vary depending on the format
    >>> pb = PinBlock(pin='1234', card_number='1111222233334444')
    >>> pb.pin
    '1234'

Pinblock encryption
^^^^^^^^^^^^^^^^^^^

The encryption mixin's adds pin block encrption and decryption. Adds **from_enc_bytes** constructor and
**to_enc_bytes** method.

.. note:: The use of an encryption mix-in requires the install of additional modules.
          Use ``pip install cardutil[crypto]``.

Available pinblock encryption mix-ins:

* :py:mod:`cardutil.pinblock.TdesEncryptedPinBlockMixin`
* :py:mod:`cardutil.pinblock.AesEncryptedPinBlockMixin`

How to encrypt and decrypt a pinblock::

    # output encrypted pinblock bytes
    >>> epb = pb.to_enc_bytes(key='00' * 16)
    >>> binascii.hexlify(epb)
    b'4c0906d10308871a'

    # create new pinblock from encrypted pinblock bytes
    >>> pb2 = PinBlock.from_enc_bytes(
    ...       enc_pin_block=epb,
    ...       card_number='1111222233334444',
    ...       key='00' * 16)
    >>> pb2.pin
    '1234'

Pin verification values
^^^^^^^^^^^^^^^^^^^^^^^

The pin verification mixin's add pin verification value calculators to the pinblock object. Adds **to_pvv** method.

.. note:: The use of a pin verification mix-in requires the install of additional modules.
          Use ``pip install cardutil[crypto]``.

Available pin verification mix-ins:

* :py:mod:`cardutil.pinblock.VisaPVVPinBlockMixin`

How to generate pin verification value::

    >>> pb.to_pvv(pvv_key='00' * 16)
    '6264'

"""

import abc
import binascii
import secrets
import logging

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend

backend = default_backend()
LOGGER = logging.getLogger(__name__)


class AbstractPinBlock(abc.ABC):
    """
    Base PinBlock object class from which implementation will subclass
    """

    def __init__(self, pin: str, *args: any, **kwargs: any):
        """
        Create a pinblock

        :param pin: string containing pin
        """
        self._pin = pin

    @property
    def pin(self) -> str:
        """
        The card pin as a string
        """
        return self._pin

    @classmethod
    @abc.abstractmethod
    def from_bytes(cls, pin_block: bytes, *args: any, **kwargs: any):
        """
        Create pinblock object from pinblock bytes

        :param pin_block: bytes containing pin block
        :return: PinBlock object
        """
        pass

    @abc.abstractmethod
    def to_bytes(self) -> bytes:
        """
        Get the pinblock bytes

        :return: bytes containing pinblock
        """
        pass


class Iso0PinBlock(AbstractPinBlock):
    """
    ISO 9564-1 format 0, ANSI X9.8, Visa-1 and ECI-0

    Pin block structure::

        P1 = LLPPPPFFFFFFFFFF
        P2 = 0000CCCCCCCCCCCC
        PIN Block = P1 XOR P2

    where::

        * L = Length of pin
        * P = Pin
        * F = Fill, x'F'
        * C = Last 12 digits of card number (excluding check digit)
    """
    def __init__(self, pin: str, card_number: str = None, **kwargs: any):
        super().__init__(pin, **kwargs)
        self.card_number = card_number

    @classmethod
    def from_bytes(cls, pin_block: bytes, card_number: str = None, *args: any, **kwargs: any):
        """
        Create object from pin block bytes

        :param pin_block: the pin block bytes
        :param card_number: the card number
        :return: the pin
        """
        rightmost_12 = card_number[-13:-1]
        p2 = f'0000{rightmost_12}'
        p1_bytes = int.from_bytes(pin_block, byteorder='big') ^ int(p2, 16)
        p1 = f'{p1_bytes:016x}'
        pin_length = int(p1[1:2])
        pin = p1[2:2 + pin_length]
        return cls(pin, card_number=card_number)

    def to_bytes(self) -> bytes:
        """
        Get the pin block bytes

        :return: pin block as bytes
        """
        rightmost_12 = self.card_number[-13:-1]
        p1 = f'{"0" + str(len(self.pin)) + self.pin:f<16}'
        p2 = f'0000{rightmost_12}'
        pin_block = int(p1, 16) ^ int(p2, 16)
        return pin_block.to_bytes(8, byteorder='big')


class Iso4PinBlock(AbstractPinBlock):
    """
    ISO 9564-1 Format 4 pin block

    https://www.ibm.com/support/knowledgecenter/SSLTBW_2.4.0/com.ibm.zos.v2r4.csfb400/iso4_sum.htm

    This block is 16 bytes long.

    Pinblock structure::

        CLPPPPaaaaaaaaAARRRRRRRRRRRRRRRR
        441234aaaaaaaaaa837c658036105d19

    where::

        * C = type of pinblock, x'4'
        * L = Length of pin, x'4' to x'C'
        * P = Pin
        * a = additional Pin or x'A'
        * A = Fill, x'A'
        * R = Random values, x'0' to x'F'
    """
    def __init__(self, pin: str, random_value: int = None, **kwargs: any):
        super().__init__(pin, **kwargs)
        if random_value:
            self.random_value = random_value
        else:
            self.random_value = secrets.randbits(64)
            LOGGER.debug(f'random_value={self.random_value}')

    def to_bytes(self) -> bytes:
        p1 = binascii.unhexlify(f'{"4" + str(len(self.pin)) + self.pin:a<16}{self.random_value:016x}')
        return p1

    @classmethod
    def from_bytes(cls, pin_block: bytes, *args: any, **kwargs: any) -> AbstractPinBlock:
        p1 = binascii.hexlify(pin_block)
        pin_length = int(p1[1:2])
        pin = p1[2:2+pin_length]
        return cls(pin.decode())


class TdesEncryptedPinBlockMixin(abc.ABC):
    """
    Adds 3DES encryption to pin blocks
    """
    @classmethod
    @abc.abstractmethod
    def from_bytes(cls, *args, **kwargs):
        pass

    @abc.abstractmethod
    def to_bytes(self, *args, **kwargs):
        pass

    @classmethod
    def from_enc_bytes(cls, enc_pin_block: bytes, key: str, *args: any, **kwargs: any):
        """
        Create pinblock object using encrypted pinblock and card number

        :param enc_pin_block: bytes containing encrypted pin block
        :param card_number: string containing card number
        :param key: hex string containing pin protection key (PPK)
        :return: PinBlock object
        """
        pin_block = cls.decrypt(key, enc_pin_block)
        return cls.from_bytes(pin_block, *args, **kwargs)

    def to_enc_bytes(self, key: str) -> bytes:
        """
        Get the encrypted pinblock bytes

        :return: bytes containing encrypted pinblock
        """
        return self.encrypt(key, self.to_bytes())

    @staticmethod
    def encrypt(key: str, data: bytes) -> bytes:
        binary_key = binascii.unhexlify(key)
        cipher = Cipher(algorithms.TripleDES(binary_key), modes.ECB(), backend=backend)
        encryptor = cipher.encryptor()
        return encryptor.update(data) + encryptor.finalize()

    @staticmethod
    def decrypt(key: str, cipher_data: bytes) -> bytes:
        binary_key = binascii.unhexlify(key)
        cipher = Cipher(algorithms.TripleDES(binary_key), modes.ECB(), backend=backend)
        decryptor = cipher.decryptor()
        return decryptor.update(cipher_data) + decryptor.finalize()


class AESEncryptedPinBlockMixin(abc.ABC):
    """
    Adds AES encryption to pin blocks
    """
    @classmethod
    @abc.abstractmethod
    def from_bytes(cls, *args, **kwargs):
        pass

    @abc.abstractmethod
    def to_bytes(self, *args, **kwargs):
        pass

    @classmethod
    def from_enc_bytes(cls, enc_pin_block: bytes, key: str, *args: any, **kwargs: any):
        """
        Create pinblock object using encrypted pinblock and card number

        :param enc_pin_block: bytes containing encrypted pin block
        :param key: hex string containing pin protection key (PPK)
        :return: PinBlock object
        """
        pin_block = cls.decrypt(key, enc_pin_block)
        return cls.from_bytes(pin_block, *args, **kwargs)

    def to_enc_bytes(self, key: str) -> bytes:
        """
        Get the encrypted pinblock bytes

        :return: bytes containing encrypted pinblock
        """
        return self.encrypt(key, self.to_bytes())

    @staticmethod
    def encrypt(key: str, data: bytes) -> bytes:
        binary_key = binascii.unhexlify(key)
        cipher = Cipher(algorithms.AES(binary_key), modes.ECB(), backend=backend)
        encryptor = cipher.encryptor()
        return encryptor.update(data) + encryptor.finalize()

    @staticmethod
    def decrypt(key: str, cipher_data: bytes) -> bytes:
        binary_key = binascii.unhexlify(key)
        cipher = Cipher(algorithms.AES(binary_key), modes.ECB(), backend=backend)
        decryptor = cipher.decryptor()
        return decryptor.update(cipher_data) + decryptor.finalize()


class VisaPVVPinBlockMixin(abc.ABC):
    """
    Adds Visa PVV calculator to pin blocks
    """
    @property
    @abc.abstractmethod
    def pin(self) -> str:
        pass

    def to_pvv(self, pvv_key, key_index=1, card_number=None):
        """
        The algorithm generates a 4-digit PIN verification value (PVV).

        :param pvv_key: the pvv key
        :param key_index: the visa key index
        :param card_number: the card number
        :return: pvv value
        """
        if hasattr(self, 'card_number'):
            card_number = self.card_number
        if not card_number:
            raise ValueError('card_number parameter must be passed')
        return calculate_pvv(self.pin, pvv_key, key_index, card_number)


def _get_tsp(card_number, key_table_index, pin):
    rightmost_11 = card_number[-12:-1]
    return f'{rightmost_11}{key_table_index}{pin}'


def calculate_pvv(pin: str, pvv_key: str, key_index: int, card_number: str):
    """
    The algorithm generates a 4-digit PIN verification value (PVV).

    See `IBM documentation
    <https://www.ibm.com/support/knowledgecenter/en/linuxonibm/com.ibm.linux.z.wskc.doc/wskc_c_appdpvvgenalg.html>`_

    :param pin: the pin to calculate PVV for
    :param pvv_key: the pvv key as a hex formatted string
    :param key_index: the visa key index
    :param card_number: the card number
    :return: pvv value
    """
    tsp = _get_tsp(card_number, key_index, pin)
    bin_pvv_key = binascii.unhexlify(pvv_key)
    cipher = Cipher(algorithms.TripleDES(bin_pvv_key), modes.ECB(), backend=backend)
    encryptor = cipher.encryptor()
    ct = encryptor.update(binascii.unhexlify(tsp)) + encryptor.finalize()
    values_pass1 = [value for value in binascii.hexlify(ct).decode() if value.isdigit()]
    if len(values_pass1) < 4:
        values_pass2 = [str(int(value, 16) - 10) for value in binascii.hexlify(ct).decode() if value.isalpha()]
        values_pass1 += values_pass2
    return ''.join(values_pass1[0:4])


class Iso0TDESPinBlockWithVisaPVV(Iso0PinBlock, TdesEncryptedPinBlockMixin, VisaPVVPinBlockMixin):
    pass


class Iso4AESPinBlockWithVisaPVV(Iso4PinBlock, AESEncryptedPinBlockMixin, VisaPVVPinBlockMixin):
    pass


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import argparse
import logging

from cardutil import __version__
from cardutil.cli import print_banner, print_exception_details
from cardutil.cli.mideu import add_logging_arg_group, add_source_format_arg
from cardutil.mciipm import MciIpmDataError, VbsReader, VbsWriter


def cli_entry(*args):
    return cli_run(**vars(_get_cli_parser().parse_args(*args)))


def cli_run(**kwargs):

    print_banner('paramconv', kwargs)
    if not len(kwargs):
        print(__name__ + " (" + __version__ + ")")
        print("try --help for information")
        return -1

    logging.basicConfig(
        level=kwargs.get('loglevel'),
        format='%(asctime)s:%(name)s:%(lineno)s:%(levelname)s:%(message)s')

    out_filename = kwargs.get('output', kwargs['input'] + '.out')

    blocked = not kwargs.get('no1014blocking', False)
    kwargs['out_encoding'] = 'utf8'
    if kwargs.get('sourceformat', 'ebcdic') == 'ebcdic':
        in_encoding = 'cp500'
        out_encoding = 'latin1'
    else:
        in_encoding = 'latin1'
        out_encoding = 'cp500'

    try:
        with open(kwargs['input'], 'rb') as in_file, open(out_filename, 'wb') as out_file:
            mci_ipm_param_encode(
                in_file, out_file=out_file, blocked=blocked,
                in_encoding=in_encoding, out_encoding=out_encoding)
    except MciIpmDataError as err:
        print_exception_details(err)
        return -1

    print("Done!")


def _get_cli_parser():
    parser = argparse.ArgumentParser(
        description="MasterCard parameter file conversion utility ({version})".format(
            version=__version__)
    )
    parser.add_argument("input", help="MasterCard parameter file name")
    parser.add_argument("-o", "--output", help="Converted parameter file name")
    parser.add_argument("--version", action="version",
                        version="%(prog)s (" + __version__ + ")",
                        help="Get version information")

    add_source_format_arg(parser)
    add_logging_arg_group(parser)

    return parser


def mci_ipm_param_encode(in_file, out_file, in_encoding, out_encoding, blocked):
    """
    Change encoding of parameter file from one encoding format to another.

    :param in_file: input parameter file object
    :param out_file: output parameter file object
    :param in_encoding: input file encoding string
    :param out_encoding: output file encoding string
    :param blocked: 1014 blocked = True, VBS = False
    :return: None
    """
    vbs_reader = VbsReader(in_file, blocked=blocked)

    in_records = (record.decode(in_encoding) for record in vbs_reader)
    out_records = (record.encode(out_encoding) for record in in_records)

    with VbsWriter(out_file, blocked=blocked) as vbs_writer:
        vbs_writer.write_many(out_records)



"""
Parsers for ISO8583 messages.

The iso8583 module provides ISO8583 message parsing functions.
See `ISO8583 Wikipedia page <https://en.wikipedia.org/wiki/ISO_8583>`_. for more details.
Also supports Mastercard |reg| PDS field structures.

The parsing functions are modelled on the python standard json library.
The functions convert raw ISO8583 messages to python dictionaries.

Dictionary keys that represent the elements of an ISO8583 message.

* MTI - Message type indicator
* DE(1-127) - Standard fields
* PDSxxxx - Mastercard PDS fields
* TAGxxxx - ICC tag fields

Import the library::

    from cardutil.iso8583 import dumps, loads

Read an ISO8583 message returning dict::

    >>> import binascii
    >>> binary_bitmap = binascii.unhexlify('c0000000000000000000000000000000')
    >>> message_bytes = b'1144' + binary_bitmap + b'164444555566667777'
    >>> message_dict = loads(message_bytes)
    >>> message_dict
    {'MTI': '1144', 'DE2': '4444555566667777'}

Create an ISO8583 message returning bytes::

    >>> message_dict = {'MTI': '1144', 'DE2': '4444555566667777'}
    >>> message_bytes = dumps(message_dict)
    >>> message_bytes
    b'1144\\xc0\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00164444555566667777'


Add **encoding** parameter if you need different message encoding.
All `standard python encoding codecs
<https://docs.python.org/3/library/codecs.html?highlight=encode#standard-encodings>`_ are available.
Default is **latin_1**.

::

    >>> message_dict = {'MTI': '1144', 'DE2': '4444555566667777'}
    >>> message_bytes = dumps(message_dict, encoding='cp500')
    >>> message_bytes
    b'\\xf1\\xf1\\xf4\\xf4\\xc0\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\xf1\\xf6\\xf4\\xf4\\xf4\\xf4\\xf5\\xf5\\xf5\\xf5\\xf6\\xf6\\xf6\\xf6\\xf7\\xf7\\xf7\\xf7'
    >>> message_dict = loads(message_bytes, encoding='cp500')
    >>> message_dict
    {'MTI': '1144', 'DE2': '4444555566667777'}

Set **hex_bitmap** to True if you require a hex format bitmap::

    >>> message_bytes = dumps(message_dict, hex_bitmap=True)
    >>> message_bytes
    b'1144c0000000000000000000000000000000164444555566667777'
    >>> message_dict = loads(message_bytes, hex_bitmap=True)
    >>> message_dict
    {'MTI': '1144', 'DE2': '4444555566667777'}

"""
import binascii
import datetime
import decimal
import logging
import re
import struct
import sys

from cardutil import CardutilError
from cardutil.BitArray import BitArray
from cardutil.card import mask
from cardutil.config import config
from cardutil.vendor.hexdump import hexdump

LOGGER = logging.getLogger(__name__)
DEFAULT_ENCODING = 'latin_1'


class Iso8583DataError(CardutilError):
    pass


def dumps(obj: dict, encoding=None, iso_config=None, hex_bitmap=False):
    """
    Serialize obj to a ISO8583 message byte string

    :param obj: dict containing message data
    :param encoding: python text encoding scheme
    :param iso_config: iso8583 message configuration dict
    :param hex_bitmap: bitmap in hex format
    :return: byte string containing ISO8583 message

    The default usage will generate a latin_1 encoded message with a binary bitmap::

        import cardutil.iso8583
        message_dict = {'MTI': '1144', 'DE2': '4444555566667777'}
        cardutil.iso8583.dumps(message_dict)

    """
    if not encoding:
        encoding = DEFAULT_ENCODING
    if not iso_config:
        iso_config = config['bit_config']

    output = _dict_to_iso8583(obj, iso_config, encoding, hex_bitmap)
    return output


def loads(b: bytes, encoding=None, iso_config=None, hex_bitmap=False):
    """
    Deserialise b (byte string) to a python object

    :param b: bytes containing message
    :param encoding: python text encoding scheme
    :param iso_config: iso8583 message configuration dictionary
    :param hex_bitmap: bitmap in hex format
    :return: dict containing message data

    ::

        import cardutil.iso8583
        message_bytes = b'1144... iso message ...'
        cardutil.iso8583.loads(message_bytes)

    """
    if not encoding:
        encoding = DEFAULT_ENCODING
    if not iso_config:
        iso_config = config['bit_config']

    return _iso8583_to_dict(b, iso_config, encoding, hex_bitmap)


def _iso8583_to_dict(message, bit_config, encoding=DEFAULT_ENCODING, hex_bitmap=False):
    """
    Convert ISO8583 style message to dictionary

    :param message: The message in ISO8583 based format

    * Message Type indicator - 4 bytes
    * Binary bitmap - 16 bytes (Reads DE1 and DE2)
    OR if hex_bitmap = True
    * Message Type indicator - 4 bytes
    * Hex bitmap - 32 bytes (Reads DE1 and DE2)
    * Message data - Remainder of record

    :param bit_config: dictionary of bit mapping configuration
    :param encoding: data encoding
    :return: dictionary of message elements

    * key = 'MTI' message type indicator
    * key = 'DEx' data elements
    * key = 'PDSxxxx' private data fields
    * key = 'TAGxxxx' icc fields

    """
    LOGGER.debug("Processing message: len=%s contents:\n%s", len(message), hexdump(message, result="return"))
    # split raw message into components MessageType(4B), Bitmap(16B),
    # Message(l=*)

    try:
        if hex_bitmap:
            message_length = len(message)-36
            message_type_indicator, bitmap, message_data = struct.unpack(
                "4s32s" + str(message_length) + "s", message)
            binary_bitmap = binascii.unhexlify(bitmap)

        else:
            message_length = len(message)-20
            message_type_indicator, binary_bitmap, message_data = struct.unpack(
                "4s16s" + str(message_length) + "s", message)
    except struct.error as ex:
        raise Iso8583DataError('Failed unpacking bitmap values', binary_context_data=message, original_exception=ex)
    return_values = dict()

    # add the message type
    try:
        return_values["MTI"] = message_type_indicator.decode(encoding)
    except UnicodeError as ex:
        raise Iso8583DataError('Failed decoding MTI field', binary_context_data=message, original_exception=ex)

    message_pointer = 0
    bitmap_list = _get_bitmap_list(binary_bitmap)

    for bit in range(2, 128):
        if bitmap_list[bit]:
            LOGGER.debug("processing bit %s", bit)
            return_message, message_increment = _iso8583_to_field(
                bit,
                bit_config[str(bit)],
                message_data[message_pointer:],
                encoding)

            # Increment the message pointer and process next field
            message_pointer += message_increment
            return_values.update(return_message)

    # check that all of message has been consumed, otherwise raise exception
    if message_pointer != len(message_data):
        raise Iso8583DataError(
            f'Message data not correct length. '
            f'Bitmap indicates len={message_pointer}, message is len={len(message_data)}',
            binary_context_data=message
        )

    return return_values


def _dict_to_iso8583(message, bit_config, encoding=DEFAULT_ENCODING, hex_bitmap=False):
    """
    Convert dictionary to ISO8583 message

    :param message: dictionary of message elements
    * key = 'MTI' message type indicator
    * key = 'DE(1-127)' data elements
    * key = 'PDSxxxx' private data fields
    * key = 'TAGxxxx' icc fields
    :param bit_config: dictionary of bit mapping configuration
    :param encoding: string indicating encoding of data
    :return: The message in ISO8583 based format
    * Message Type indicator - 4 bytes
    * Binary bitmap - 16 bytes (Reads DE1)
    * Message data - Remainder of record

    """
    output_data = b''
    bitmap_values = [False] * 128
    bitmap_values[0] = True  # set bit 1 on for presence of bitmap

    # get the pds fields from config
    de_pds_fields = sorted(
        [int(key) for key in bit_config if bit_config[key].get('field_processor') == 'PDS'], reverse=True)
    LOGGER.debug(de_pds_fields)

    for de_field_value in _pds_to_de(message):
        de_field_key = de_pds_fields.pop()
        LOGGER.debug(f'de{de_field_key}={de_field_value}')
        message[f'DE{de_field_key}'] = de_field_value

    for bit in range(2, 128):
        if message.get('DE' + str(bit)) or message.get('DE' + str(bit)) == 0:  # 0 evals to false, allow zero values
            LOGGER.debug(f'processing bit {bit}')
            bitmap_values[bit - 1] = True
            LOGGER.debug(message.get('DE' + str(bit)))
            output_data += _field_to_iso8583(
                bit_config[str(bit)],
                message.get('DE' + str(bit)),
                encoding=encoding)

    bitarray = BitArray()
    bitarray.fromlist(bitmap_values)
    binary_bitmap = bitarray.tobytes()
    if hex_bitmap:
        bitmap = binascii.hexlify(binary_bitmap)
    else:
        bitmap = binary_bitmap

    mti = message['MTI'].encode(encoding) if message.get('MTI') else b''
    output_string = mti + bitmap + output_data
    return output_string


def _field_to_iso8583(bit_config, field_value, encoding=DEFAULT_ENCODING):

    output = ''
    field_value = _pytype_to_string(field_value, bit_config)
    field_length = bit_config.get('field_length')
    length_size = _get_field_length(bit_config)  # size of length for llvar and lllvar fields
    if length_size > 0:
        field_length = len(field_value)
        output += format(field_length, '0' + str(length_size))
    output += format(field_value[:field_length], '<' + str(field_length))
    return output.encode(encoding)


def _iso8583_to_field(bit, bit_config, message_data, encoding=DEFAULT_ENCODING):
    """
    Processes a message bit element

    :param bit: DE bit
    :param bit_config: message bit configuration
    :param message_data: the data to be processed
    :param encoding: byte encoding
    :returns:
        dictionary: field values
        message incrementer: position of next message
    """

    field_length = bit_config['field_length']

    length_size = _get_field_length(bit_config)

    if length_size > 0:
        field_length_string = message_data[:length_size]
        try:
            field_length_string = field_length_string.decode(encoding)
        except UnicodeDecodeError as ex:
            raise Iso8583DataError(f'Unable to decode DE{bit} field length',
                                   binary_context_data=message_data, original_exception=ex)

        try:
            field_length = int(field_length_string)
        except ValueError as ex:
            raise Iso8583DataError(f'Invalid field length DE{bit}',
                                   binary_context_data=message_data, original_exception=ex)

    field_data = message_data[length_size:length_size + field_length]
    LOGGER.debug(f'field_data={field_data}')
    field_processor = bit_config.get('field_processor')

    # do ascii conversion except for ICC field
    if field_processor != 'ICC':
        try:
            field_data = field_data.decode(encoding)
        except UnicodeDecodeError as ex:
            raise Iso8583DataError(f'Unable to decode DE{bit} field value',
                                   binary_context_data=message_data, original_exception=ex)

    # if field is PAN type, mask the card value
    if field_processor == 'PAN':
        field_data = mask(field_data)

    # if field is PAN type, mask the card value
    if field_processor == 'PAN-PREFIX':
        field_data = _pan_prefix(field_data)

    # do field conversion to native python type
    try:
        field_data = _string_to_pytype(field_data, bit_config)
    except ValueError as ex:
        raise Iso8583DataError(f'Unable to convert DE{bit} field to python type',
                               binary_context_data=message_data, original_exception=ex)
    return_values = dict()

    # add value to return dictionary
    return_values["DE" + str(bit)] = field_data

    # if a PDS field, break it down again and add to results
    if field_processor == 'PDS':
        return_values.update(_pds_to_dict(field_data))

    # if a DE43 field, break in down again and add to results
    if field_processor == 'DE43':
        processor_config = bit_config.get('field_processor_config')
        return_values.update(_get_de43_fields(field_data, processor_config))

    # if ICC field, break into tags
    if field_processor == 'ICC':
        return_values.update(_icc_to_dict(field_data))

    return return_values, field_length + length_size


def _pan_prefix(field_data):
    """
    Get prefix of PAN
    """
    return field_data[:9]


def _string_to_pytype(field_data, bit_config):
    """
    Field conversion to native python type

    :param field_data: Data to be converted
    :param bit_config: Configuration for bit
    :return: data in required type
    """
    field_python_type = bit_config.get('field_python_type')
    field_date_format = bit_config.get('field_date_format', "%y%m%d")

    if field_python_type in ("int", "long"):
        field_data = int(field_data)
    if field_python_type == "decimal":
        field_data = decimal.Decimal(field_data)
    if field_python_type == "datetime":
        field_data = datetime.datetime.strptime(field_data, field_date_format)
    return field_data


def _pytype_to_string(field_data, bit_config):
    """
    convert py type to string for message

    :param field_data: Data to be converted
    :param bit_config: Configuration for bit
    :return: data in required type
    """
    field_python_type = bit_config.get('field_python_type')
    field_date_format = bit_config.get('field_date_format', "%y%m%d")

    return_string = field_data
    if field_python_type in ('int', 'long'):
        return_string = format(int(field_data), '0' + str(bit_config.get('field_length', 0)) + 'd')
    if field_python_type == "decimal":
        return_string = format(decimal.Decimal(field_data), '0' + str(bit_config.get('field_length', 0)) + 'f')
    if field_python_type == "datetime":
        if not isinstance(field_data, datetime.datetime):
            field_data = _get_date_from_string(field_data)
        return_string = format(field_data, field_date_format)
    return return_string


def _get_date_from_string(field_data: str) -> datetime:
    """
    Parse string dates to python datetime object

    Use dateutils library if it is installed, otherwise revert to simple parser
    :param field_data: string containing date
    :return: datetime object
    """
    try:
        import dateutil.parser as parser
        print('Using dateutil parser')
        return parser.parse(field_data)
    except ImportError:
        pass

    if sys.version_info >= (3, 7):
        print('Using fromisoformat')
        return datetime.datetime.fromisoformat(field_data)

    # fallback parser -- tries a few different formats until one works
    print('Using built in date parser')
    date_formats = ["%Y-%m-%d %H:%M:%S", "%Y-%m-%d %H:%M", "%Y-%m-%d"]
    output_date = None
    for date_format in date_formats:
        try:
            output_date = datetime.datetime.strptime(field_data, date_format)
            break
        except ValueError:
            continue
    if not output_date:
        raise ValueError("Unrecognised date string format - {}".format(field_data))
    return output_date


def _get_field_length(bit_config):
    """
    Determine length of iso8583 style field

    :param bit_config: dictionary of bit config data
    :return: length of field
    """
    length_size = 0

    if bit_config['field_type'] == "LLVAR":
        length_size = 2
    elif bit_config['field_type'] == "LLLVAR":
        length_size = 3

    return length_size


def _get_bitmap_list(binary_bitmap):
    """
    Get list of bits from binary bitmap

    :param binary_bitmap: the binary bitmap to be returned
    :return: the list containing bit values. Bit 0 contains original binary bitmap
    """
    working_bitmap_list = BitArray(endian='big')
    working_bitmap_list.frombytes(binary_bitmap)

    # Add bit 0 -> original binary bitmap
    bitmap_list = [binary_bitmap]

    # add bits from bitmap
    bitmap_list.extend(working_bitmap_list.tolist())

    return bitmap_list


def _pds_to_de(dict_values):
    """
    takes all the pds fields values in dict (PDSxxxx) and creates list of DE strings

    :param dict_values: dict containing "PDSxxxx" elements
    :return: list of byte strings containing pds data, or None if no fields
    """
    # get the PDS field keys in order
    LOGGER.debug(f'dict_values={dict_values}')
    keys = sorted([key for key in dict_values if key.startswith('PDS')])
    LOGGER.debug(f'keys={keys}')
    output = ''
    outputs = []
    for key in keys:
        tag = int(key[3:])
        LOGGER.debug(f'tag={tag}')
        length = len(dict_values[key])
        add_output = f'{tag:04}{length:03}{dict_values[key]}'
        if len(output + add_output) > 999:
            outputs.append(output)
            output = ''
        output += add_output
    if output:
        outputs.append(output)
    LOGGER.debug(f'>pds2de: {outputs}')

    return outputs


def _pds_to_dict(field_data):
    """
    Get MasterCard pds fields from iso field

    :param field_data: the ISO8583 field containing pds fields
    :return: dictionary of pds key values. Key in the form PDSxxxx where x is zero filled number of pds
    """
    field_pointer = 0
    return_values = {}

    while field_pointer < len(field_data):
        # get the pds tag id
        pds_field_tag = field_data[field_pointer:field_pointer+4]
        LOGGER.debug("pds_field_tag=[%s]", pds_field_tag)

        # get the pds length
        pds_field_length = int(field_data[field_pointer+4:field_pointer+7])
        LOGGER.debug("pds_field_length=[%i]", pds_field_length)

        # get the pds data
        pds_field_data = field_data[field_pointer+7:field_pointer+7+pds_field_length]
        LOGGER.debug("pds_field_data=[%s]", str(pds_field_data))
        return_values["PDS" + pds_field_tag] = pds_field_data

        # increment the fieldPointer
        field_pointer += 7+pds_field_length

    return return_values


def _icc_to_dict(field_data):
    """
    Get de55 fields from message

    :param field_data: the field containing de55
    :return: dictionary of de55 key values
             key is tag+tagid
    """
    TWO_BYTE_TAG_PREFIXES = [b'\x9f', b'\x5f']

    field_pointer = 0
    return_values = {"ICC_DATA": binascii.b2a_hex(field_data).decode()}

    while field_pointer < len(field_data):
        # get the tag id (one byte)
        field_tag = field_data[field_pointer:field_pointer+1]
        # set to 2 bytes if 2 byte tag
        if field_tag in TWO_BYTE_TAG_PREFIXES:
            field_tag = field_data[field_pointer:field_pointer+2]
            field_pointer += 2
        else:
            field_pointer += 1

        field_tag_display = binascii.b2a_hex(field_tag)
        LOGGER.debug("field_tag_display=%s", field_tag_display)

        # stop processing de55 if low values tag found
        if field_tag_display == b'00':
            break

        field_length_raw = field_data[field_pointer:field_pointer+1]
        field_length = struct.unpack(">B", field_length_raw)[0]

        LOGGER.debug("%s", format(field_tag_display))
        LOGGER.debug(field_length)

        # get the tag data
        de_field_data = field_data[field_pointer+1:field_pointer+field_length+1]
        de_field_data_display = binascii.b2a_hex(de_field_data).decode()
        LOGGER.debug("%s", de_field_data_display)
        return_values["TAG" + field_tag_display.upper().decode()] = de_field_data_display

        # increment the fieldPointer
        field_pointer += 1+field_length

    return return_values


def _get_de43_fields(de43_field, processor_config=None):
    """
    get pds 43 field breakdown

    :param de43_field: data of pds 43
    :return: dictionary of pds 43 sub elements
    """
    LOGGER.debug("de43_field=%s", de43_field)
    if processor_config:
        de43_regex = processor_config
    else:
        de43_regex = (
            r"(?P<DE43_NAME>.+?) *\\(?P<DE43_ADDRESS>.+?) *\\(?P<DE43_SUBURB>.+?) *\\"
            r"(?P<DE43_POSTCODE>\S{4,10}) *(?P<DE43_STATE>.{3})(?P<DE43_COUNTRY>.{3})"
        )

    field_match = re.match(de43_regex, de43_field)
    if not field_match:
        return dict()

    # get the dict
    field_dict = field_match.groupdict()

    return field_dict


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import io
import os
import tempfile
import unittest
from tests import message_ascii_raw, print_stream

from cardutil.mciipm import VbsWriter
from cardutil.cli import mci_ipm_encode


class MciIpmEncodeTestCase(unittest.TestCase):
    def test_mci_ipm_encode(self):
        # create test ipm file
        message_list = [message_ascii_raw for _ in range(5)]
        vbs_in = io.BytesIO()
        writer = VbsWriter(vbs_in, blocked=True)
        for message in message_list:
            writer.write(message)
        writer.close()

        # process the ipm encode
        ipm_out = io.BytesIO()
        mci_ipm_encode.mci_ipm_encode(vbs_in, ipm_out, in_encoding='ascii', out_encoding='ascii')

        print_stream(vbs_in, "Input")
        vbs_in_value = vbs_in.read()

        print_stream(ipm_out, "Output")
        ipm_out_value = ipm_out.read()

        self.assertEqual(vbs_in_value, ipm_out_value)

    def test_mci_ipm_encode_cli_parser(self):
        args = vars(mci_ipm_encode.cli_parser().parse_args(['file1.ipm']))
        self.assertEqual(
            args,
            {'in_filename': 'file1.ipm', 'out_filename': None, 'in_encoding': None,
             'out_encoding': None, 'in_format': '1014', 'out_format': '1014',
             'no1014blocking': False, 'debug': False})

    def test_mci_ipm_encode_input_params(self):
        """
        Run mci_ipm_encode using real files
        """
        # create an ipm file
        with tempfile.NamedTemporaryFile(mode='wb', delete=False) as out_ipm:
            out_ipm.write(
                b'\x00\x00\x00\x1a0100\x80\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                b'n\x9cm\x9cl\x9c\x00\x00\x00\x00')
            in_ipm_name = out_ipm.name
            out_ipm.close()

        mci_ipm_encode.cli_run(in_filename=in_ipm_name)
        mci_ipm_encode.cli_run(in_filename=in_ipm_name, out_filename=in_ipm_name + '.out')
        mci_ipm_encode.cli_run(in_filename=in_ipm_name, no1014blocking=True)
        mci_ipm_encode.cli_run(in_filename=in_ipm_name, debug=True)
        os.remove(in_ipm_name)
        os.remove(in_ipm_name + '.out')



import argparse
import logging
from csv import DictReader

from cardutil.cli import add_version, get_config, print_banner
from cardutil.mciipm import IpmWriter


def cli_entry():
    cli_run(**vars(cli_parser().parse_args()))


def cli_run(**kwargs):
    print_banner('mci_csv_to_ipm', kwargs)

    if kwargs.get('debug'):
        logging.basicConfig(level=logging.DEBUG)

    config = get_config('cardutil.json', cli_filename=kwargs.get('config_file'))

    if not kwargs.get('out_filename'):
        kwargs['out_filename'] = kwargs['in_filename'] + '.ipm'

    with open(kwargs['in_filename'], 'r', encoding=kwargs.get('in_encoding')) as in_csv:
        with open(kwargs['out_filename'], 'wb') as out_ipm:
            mci_csv_to_ipm(in_csv=in_csv, out_ipm=out_ipm, config=config, **kwargs)


def cli_parser():
    parser = argparse.ArgumentParser(prog='mci_csv_to_ipm', description='CSV to Mastercard IPM')
    parser.add_argument('in_filename')
    parser.add_argument('-o', '--out-filename')
    parser.add_argument('--in-encoding')
    parser.add_argument('--out-encoding')
    parser.add_argument('--no1014blocking', action='store_true')
    parser.add_argument('--config-file', help='File containing cardutil configuration - JSON format')
    parser.add_argument('--debug', action='store_true')
    add_version(parser)

    return parser


def mci_csv_to_ipm(in_csv, out_ipm, config, out_encoding=None, no1014blocking=False, **_):
    """
    Create a Mastercard IPM file given an input csv file object

    :param in_csv: file object containing csv formatted data
    :param out_ipm: file object to receive IPM file
    :param config: dict containing cardutil config
    :param out_encoding: The IPM file encoding. 'cp500' works for EBCDIC
    :param no1014blocking: set True 1014 blocking not required
    :return: None
    """
    blocked = not no1014blocking
    with IpmWriter(out_ipm, encoding=out_encoding, blocked=blocked, iso_config=config.get('bit_config')) as writer:
        reader = DictReader(in_csv)
        writer.write_many(reader)


if __name__ == '__main__':
    cli_entry()



import argparse
import logging

from cardutil.cli import add_version, print_banner
from cardutil.mciipm import VbsReader, VbsWriter


def cli_entry():
    cli_run(**vars(cli_parser().parse_args()))


def cli_run(**kwargs):
    print_banner('mci_ipm_param_encode', kwargs)

    if kwargs.get('debug'):
        logging.basicConfig(level=logging.DEBUG)

    if not kwargs.get('out_filename'):
        kwargs['out_filename'] = kwargs['in_filename'] + '.out'

    # if --no1014encoding provided, override new in-format and out-format parms
    if kwargs.get('no1014blocking'):
        kwargs['in_format'] = kwargs['out_format'] = 'vbs'

    with open(kwargs['in_filename'], 'rb') as in_file, open(kwargs['out_filename'], 'wb') as out_file:
        mci_ipm_param_encode(in_file, out_file=out_file, **kwargs)


def cli_parser():

    parser = argparse.ArgumentParser(prog='mci_ipm_param_encode', description='Mastercard IPM param file encoder')
    parser.add_argument('in_filename')
    parser.add_argument('-o', '--out-filename')
    parser.add_argument('--in-encoding')
    parser.add_argument('--out-encoding')
    parser.add_argument('--no1014blocking', action='store_true')
    parser.add_argument('--in-format', choices=['vbs', '1014'], default='1014')
    parser.add_argument('--out-format', choices=['vbs', '1014'], default='1014')
    parser.add_argument('--debug', action='store_true')
    add_version(parser)

    return parser


def mci_ipm_param_encode(in_file, out_file, in_encoding=None, out_encoding=None,
                         in_format='1014', out_format='1014', **_):
    """
    Change encoding of parameter file from one encoding format to another.

    :param in_file: input parameter file object
    :param out_file: output parameter file object
    :param in_encoding: input file encoding string
    :param out_encoding: output file encoding string
    :param in_format: input file format (vbs/1014)
    :param out_format: output file format (vbs/1014)
    :return: None
    """
    if not in_encoding:
        in_encoding = 'latin_1'
    if not out_encoding:
        out_encoding = in_encoding

    in_blocked = True if in_format == '1014' else False
    vbs_reader = VbsReader(in_file, blocked=in_blocked)

    in_records = (record.decode(in_encoding) for record in vbs_reader)
    out_records = (record.encode(out_encoding) for record in in_records)
    out_blocked = True if out_format == '1014' else False
    with VbsWriter(out_file, blocked=out_blocked) as vbs_writer:
        vbs_writer.write_many(out_records)


if __name__ == '__main__':
    cli_entry()



import unittest

from cardutil import BitArray


class BitArrayTestCase(unittest.TestCase):
    def test_big_endian(self):
        my_array = BitArray.BitArray()
        my_array.frombytes(b"\x01")
        self.assertEqual([False, False, False, False, False, False, False, True], my_array.tolist())

    def test_liitle_endian(self):
        my_array = BitArray.BitArray(endian='little')
        my_array.frombytes(b"\x01")
        self.assertEqual([True, False, False, False, False, False, False, False], my_array.tolist())


if __name__ == '__main__':
    unittest.main()



import unittest

from cardutil import pinblock


class PinblockTestCase(unittest.TestCase):
    def test_pin_block_Iso0(self):
        self.assertEqual(
            b'\x04\x12\x26\xcb\xa9\x87\x6f\xed',
            pinblock.Iso0PinBlock(pin='1234', card_number='4441234567890123').to_bytes())

    def test_visa_pvv(self):
        test_key = '5CA64B3C22BEC347CA7E6609904BAAED'
        self.assertEqual(
            '3856', pinblock.calculate_pvv(pin='2205', card_number='4564320000980369', pvv_key=test_key, key_index=1))

    def test_visa_pvv_more_values_required(self):
        """
        check when additional digits for pvv required
        """
        test_key = '5CA64B3C22BEC347CA7E6609904BAAED'
        self.assertEqual(
            '0885', pinblock.calculate_pvv(pin='0654', card_number='4564320000980369', pvv_key=test_key, key_index=1))

    def test_visa_pvv_mixin(self):
        # use pin block without card_number property
        MyPinBlock = type('MyPinBlock', (pinblock.Iso4PinBlock, pinblock.VisaPVVPinBlockMixin), {})
        pb = MyPinBlock(pin='6666')
        with self.assertRaises(ValueError):
            pb.to_pvv(pvv_key='00' * 8)
        self.assertEqual(pb.to_pvv(pvv_key='00' * 8, card_number='1111222233334444'), '1703')

    def test_pin_block_Iso0TDESPinBlockWithVisaPVV(self):
        pb1 = pinblock.Iso0TDESPinBlockWithVisaPVV(pin='1234', card_number='1111222233334444')
        self.assertEqual(pb1.pin, '1234')
        self.assertEqual(pb1.card_number, '1111222233334444')
        self.assertEqual(pb1.to_bytes(), b'\x04\x12&\xdd\xdc\xcc\xcb\xbb')
        self.assertEqual(pb1.to_enc_bytes(key='00' * 16), b'L\t\x06\xd1\x03\x08\x87\x1a')
        self.assertEqual(pb1.to_pvv(pvv_key='00' * 16), '6264')

        pb2 = pinblock.Iso0TDESPinBlockWithVisaPVV.from_bytes(
            pin_block=pb1.to_bytes(), card_number='1111222233334444')
        self.assertEqual(pb2.pin, '1234')

        pb3 = pinblock.Iso0TDESPinBlockWithVisaPVV.from_enc_bytes(
            enc_pin_block=pb1.to_enc_bytes('00' * 16), card_number='1111222233334444', key='00' * 16)
        self.assertEqual(pb3.pin, '1234')

    def test_pin_block_Iso4AESPinBlockWithVisaPVV(self):
        pb1 = pinblock.Iso4AESPinBlockWithVisaPVV(pin='1234', random_value=14932500169729639426)
        self.assertEqual(pb1.pin, '1234')
        self.assertEqual(pb1.to_bytes()[0:8], b'\x44\x12\x34\xaa\xaa\xaa\xaa\xaa')
        self.assertEqual(pb1.to_enc_bytes(key='00' * 16), b',4yaY\xbf\x10j\xf6\xf5\xd2;Y\xfd\xe2\xfe')
        self.assertEqual(pb1.to_pvv(pvv_key='00' * 16, card_number='1111222233334444'), '6264')

        pb2 = pinblock.Iso4AESPinBlockWithVisaPVV.from_bytes(pin_block=pb1.to_bytes())
        self.assertEqual(pb2.pin, '1234')

        pb3 = pinblock.Iso4AESPinBlockWithVisaPVV.from_enc_bytes(
            enc_pin_block=pb1.to_enc_bytes('00' * 16), key='00' * 16)
        self.assertEqual(pb3.pin, '1234')

    def test_pinblock_operations(self):
        tdes_key = '2222111122221111'
        aes_key = tdes_key * 2
        key = aes_key

        pin_blocks = [
            type(
                'TDESPinBlock',
                (pinblock.Iso0PinBlock, pinblock.TdesEncryptedPinBlockMixin, pinblock.VisaPVVPinBlockMixin), {}),
            type(
                'AESPinBlock',
                (pinblock.Iso4PinBlock, pinblock.AESEncryptedPinBlockMixin, pinblock.VisaPVVPinBlockMixin), {}),
            pinblock.Iso0TDESPinBlockWithVisaPVV,
            pinblock.Iso4AESPinBlockWithVisaPVV
        ]
        for pb in pin_blocks:
            print(pb)
            pb = pb(pin='1234', card_number='1111222233334444')
            self.assertEqual(pb.pin, '1234')
            self.assertEqual(pb.to_pvv(pvv_key='1111222211112222', key_index=1, card_number='1111222233334444'), '9595')
            pb2 = pb.from_bytes(pb.to_bytes(), card_number='1111222233334444')
            self.assertEqual(pb2.pin, '1234')
            pb2 = pb.from_enc_bytes(pb.to_enc_bytes(key=key), key=key, card_number='1111222233334444')
            self.assertEqual(pb2.pin, '1234')


if __name__ == '__main__':
    unittest.main()



import argparse
import collections
import csv
import logging

from cardutil.cli import add_version, get_config, print_banner, print_exception_details
from cardutil.mciipm import IpmReader, MciIpmDataError


def cli_entry():
    return cli_run(**vars(cli_parser().parse_args()))


def cli_run(**kwargs):

    print_banner('mci_ipm_to_csv', kwargs)

    if kwargs.get('debug'):
        logging.basicConfig(level=logging.DEBUG)

    config = get_config('cardutil.json', cli_filename=kwargs.get('config_file'))

    if not kwargs.get('out_filename'):
        kwargs['out_filename'] = kwargs['in_filename'] + '.csv'

    try:
        with open(kwargs['in_filename'], 'rb') as in_ipm:
            with open(kwargs['out_filename'], 'w', encoding=kwargs.get('out_encoding')) as out_csv:
                mci_ipm_to_csv(in_ipm=in_ipm, out_csv=out_csv, config=config, **kwargs)
    except MciIpmDataError as err:
        print_exception_details(err)
        return -1


def dicts_to_csv(data_list, output_file, field_list=None):
    """
    Writes dict data to CSV file

    :param data_list: list of dictionaries that contain the data to be loaded
    :param output_file: output CSV file
    :param field_list: (optional) list of fields to output to CSV file
    :return: None
    """

    if not field_list:
        # get the fields present in the dicts
        field_summary = collections.Counter()
        data_list = list(data_list)
        for item in data_list:
            field_summary.update(item.keys())
        field_list = [field_key for field_key in field_summary]

    writer = csv.DictWriter(
        output_file,
        fieldnames=field_list,
        extrasaction="ignore",
        lineterminator="\n")

    writer.writeheader()
    for data_item in data_list:
        writer.writerow({item: data_item[item] for item in data_item if item in field_list})


def cli_parser():
    parser = argparse.ArgumentParser(prog='mci_ipm_to_csv', description='Mastercard IPM to CSV')
    parser.add_argument('in_filename')
    parser.add_argument('-o', '--out-filename')
    parser.add_argument('--in-encoding')
    parser.add_argument('--out-encoding')
    parser.add_argument('--no1014blocking', action='store_true')
    parser.add_argument('--config-file', help='File containing cardutil configuration - JSON format')
    parser.add_argument('--debug', action='store_true')
    add_version(parser)

    return parser


def mci_ipm_to_csv(in_ipm, out_csv, config, in_encoding=None, no1014blocking=False, **_):
    """
    Create a csv file given an input Mastercard IPM file

    :param in_ipm: binary input IPM file object
    :param out_csv: output csv file object
    :param config: dict containing cardutil config
    :param in_encoding: input file encoding string
    :param no1014blocking: set True if no 1014 blocking used
    :return: None
    """
    blocked = not no1014blocking
    dicts_to_csv(
        IpmReader(in_ipm, encoding=in_encoding, blocked=blocked, iso_config=config.get('bit_config')),
        out_csv, field_list=config.get('output_data_elements'))



import binascii
import unittest


from cardutil.key import get_zone_master_key, get_enc_zone_master_key, calculate_kcv


class KeyTestCase(unittest.TestCase):
    def test_get_zone_master_key(self):
        hex_key, kcv = get_zone_master_key('6d6be51f04f76167491554fe25f7abef', '67499b2cf137dfcb9ea28ff757cd10a7')
        self.assertEqual('0a227e33f5c0beacd7b7db09723abb48', hex_key)
        self.assertEqual('05ee1d', kcv)

    def test_get_enc_zone_master_key(self):
        enc_key, kvc = get_enc_zone_master_key(
            '00' * 16,
            '6d6be51f04f76167491554fe25f7abef', '67499b2cf137dfcb9ea28ff757cd10a7')
        self.assertEqual(enc_key, '06ed6dbd8e7d3a9431a9df6ab329df3e')
        self.assertEqual(kvc, '05ee1d')

    def test_calculate_kvc(self):
        self.assertEqual(calculate_kcv(binascii.unhexlify('67C4A7191ADAFD086432CE0DD6384AB8')), '20d40b')
        self.assertEqual(
            calculate_kcv(binascii.unhexlify('67C4A7191ADAFD086432CE0DD6384AB8'), kvc_length=8), '20d40bfb')


if __name__ == '__main__':
    unittest.main()



import unittest
import io
import contextlib

from cardutil.cli import print_banner, print_exception_details
from cardutil.mciipm import MciIpmDataError


class GetConfigTestCase(unittest.TestCase):

    def test_print_banner(self):
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            print_banner("test_command", {'parm1': 1, 'parm2': None})
        output = f.getvalue().splitlines()
        self.assertRegex(output[0], r'test_command \(cardutil \d*\.\d*\.\d*\)')  # don't include version
        assert output[1] == 'parameters:'
        assert output[2] == ' -parm1:1'

    def test_print_mciipmdataerror(self):
        ex = MciIpmDataError("A data error", record_number=1, binary_context_data=b'1234')
        print_exception_details(ex)
        ex = MciIpmDataError("A data error")  # plain error
        print_exception_details(ex)



import json
import logging
import os

from cardutil.config import config as pkg_config
from cardutil import __version__, CardutilError
from cardutil.vendor.hexdump import hexdump

LOGGER = logging.getLogger(__name__)


def print_banner(command_name, parms):
    message = f'{command_name} (cardutil {__version__})'
    print(message)
    print('parameters:')
    for parm_key in parms:
        if parms[parm_key]:
            print(f' -{parm_key}:{parms[parm_key]}')


def print_exception_details(err: CardutilError):
    print("*** ERROR - processing has stopped ***")
    if err.record_number:
        print(f'Error detected in record {err.record_number}')
    print(err)
    if err.binary_context_data:
        hexdump(err.binary_context_data)


def add_version(parser):
    version_text = (f'%(prog)s (cardutil {__version__})\n'
                    f'(C)Copyright 2019-2022 Anthony Delosa\n')

    parser.add_argument('--version', action='version', version=version_text)


def get_config(config_filename, envvar='CARDUTIL_CONFIG', cli_filename=None):
    # check for json config file provided from command line argument
    if cli_filename:
        if os.path.isfile(cli_filename):
            LOGGER.info('Using cli config at {}'.format(cli_filename))
            with open(cli_filename, 'r') as f:
                config = f.read()
            return json.loads(config)

    # check for json config file from ENVVAR directory
    config_dir = os.environ.get(envvar)
    LOGGER.debug('config_dir={}'.format(config_dir))
    if config_dir:
        config_dir = os.path.abspath(config_dir)
        LOGGER.debug(os.path.join(config_dir, config_filename))
        if os.path.isfile(os.path.join(config_dir, config_filename)):
            LOGGER.info('Using config at {}'.format(config_dir))
            config_filename = os.path.join(config_dir, config_filename)
            with open(config_filename, 'r') as f:
                config = f.read()
            return json.loads(config)

    # Use package config
    return pkg_config



"""
mideu is a command line utility that is part of a deprecated package mciutil
The goal is for this to be compatible with that tool while using the new
engine from cardutil.
"""
import os
import tempfile
import unittest

from cardutil.cli import mideu

CONFIG_DATA = """
{
    "bit_config": {
        "38": {
            "field_name": "Approval code",
            "field_type": "FIXED",
            "field_length": 6
        }
    }
}
"""


class MideuTestCases(unittest.TestCase):
    def test_mideu_no_params(self):
        mideu.cli_run()

    def test_mideu_help(self):
        with self.assertRaises(SystemExit) as ex:
            mideu.cli_entry(['--help'])
        self.assertEqual(ex.exception.code, 0)

    def test_mideu_extract_exception(self):
        in_ipm_data = (b'\x00\x00\x00\x1a0100\x80\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                       b'nXmXlX\xFF\xFF\x00\x00')

        with tempfile.NamedTemporaryFile(mode='w+b', delete=False) as in_ipm:
            in_ipm.write(in_ipm_data)
            in_ipm_name = in_ipm.name
            print(in_ipm_name)
            in_ipm.close()

            mideu.cli_run(
                input=in_ipm_name,
                csvoutputfile=in_ipm_name + '.csv',
                sourceformat='ascii',
                debug=True,
                func=mideu.extract
            )

        os.remove(in_ipm_name)
        os.remove(in_ipm_name + '.csv')

    def test_mideu_extract(self):
        """
        Runs mideu extract command from cli_run
        """
        in_ipm_data = (b'\x00\x00\x00\x1a0100\x80\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                       b'nXmXlX\x00\x00\x00\x00')

        with tempfile.NamedTemporaryFile(mode='w+b', delete=False) as in_ipm:
            in_ipm.write(in_ipm_data)
            in_ipm_name = in_ipm.name
            print(in_ipm_name)
            in_ipm.close()

        # run with config file
        with tempfile.NamedTemporaryFile(mode='w+', delete=False) as in_config:
            config_filename = in_config.name
            in_config.write(CONFIG_DATA)
            in_config.close()
            mideu.cli_run(
                config_file=config_filename,
                input=in_ipm_name,
                # csvoutputfile=in_ipm_name + '.csv',
                sourceformat='ascii',
                debug=True,
                func=mideu.extract
            )
            os.remove(config_filename)

        csv_output = open(in_ipm_name + '.csv', 'r').read()
        self.assertEqual(csv_output, "MTI,DE38\n0100,nXmXlX\n")
        os.remove(in_ipm_name)
        os.remove(in_ipm_name + '.csv')

    def test_mideu_extract_with_output(self):
        """
        Runs mideu extract command from cli_run, output file specified
        """
        in_ipm_data = (b'\x00\x00\x00\x1a0100\x80\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                       b'nXmXlX\x00\x00\x00\x00')

        with tempfile.NamedTemporaryFile(mode='w+b', delete=False) as in_ipm:
            in_ipm.write(in_ipm_data)
            in_ipm_name = in_ipm.name
            print(in_ipm_name)
            in_ipm.close()

        # run with config file
        with tempfile.NamedTemporaryFile(mode='w+', delete=False) as in_config:
            config_filename = in_config.name
            in_config.write(CONFIG_DATA)
            in_config.close()
            mideu.cli_run(
                config_file=config_filename,
                input=in_ipm_name,
                csvoutputfile=in_ipm_name + '.csv',
                sourceformat='ascii',
                debug=True,
                func=mideu.extract
            )
            os.remove(config_filename)

        csv_output = open(in_ipm_name + '.csv', 'r').read()
        self.assertEqual(csv_output, "MTI,DE38\n0100,nXmXlX\n")
        os.remove(in_ipm_name)
        os.remove(in_ipm_name + '.csv')

    def test_mideu_convert(self):
        """
        Run mideu convert from cli_run
        """
        # create an ipm file
        with tempfile.NamedTemporaryFile(mode='wb', delete=False) as out_ipm:
            out_ipm.write(
                b'\x00\x00\x00\x1a0100\x80\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                b'n\x9cm\x9cl\x9c\x00\x00\x00\x00')
            in_ipm_name = out_ipm.name
            out_ipm.close()

        mideu.cli_run(func=mideu.convert, input=in_ipm_name)
        mideu.cli_run(func=mideu.convert, input=in_ipm_name, no1014blocking=True)
        mideu.cli_run(func=mideu.convert, input=in_ipm_name + '.out', no1014blocking=True, sourceformat='ascii')
        mideu.cli_run(func=mideu.extract, input=in_ipm_name + '.out', no1014blocking=True)

        os.remove(in_ipm_name)
        os.remove(in_ipm_name + '.out')
        os.remove(in_ipm_name + '.out' + '.csv')
        os.remove(in_ipm_name + '.out' + '.out')
